from typing import Dict, List

from .name2idx import C, V


class ReactionNetwork(object):
    def __init__(self) -> None:
        """
        Reaction indices grouped according to biological processes.
        This is used for sensitivity analysis (target='reaction').
        """
        super(ReactionNetwork, self).__init__()

        self.reactions: Dict[str, List[int]] = {
            "all": [i for i in range(1, 114)],
        }

    @staticmethod
    def flux(t, y, x) -> dict:
        """
        Rate equations in the model.

        Parameters
        ----------
        t : float
            Time point.
        y : ndarray
            Concentration vector.
        x : tuple
            Parameter values.

        Returns
        -------
        v : dict
            Flux vector.
        """
        v = {}
        # 1 ∅ −→ EGFR
        v[1] = (
            2
            * x[C.EGFR_basal_activation]
            * (x[C.init_EGFR] ** 2)
            * x[C.pEGFR_degradation]
            / (x[C.pEGFR_degradation] + x[C.init_RTKph] * x[C.pEGFR_phosphatase_binding])
            + x[C.EGFR_ErbB2_basal_act]
            * x[C.init_EGFR]
            * x[C.pErbB12_degradation]
            / (x[C.pErbB12_degradation] + x[C.init_RTKph] * x[C.pErbB12i_phosphatase])
            + x[C.EGFR_ErbB3_basal_act]
            * x[C.init_EGFR]
            * x[C.init_ErbB3]
            * x[C.pErbB13_degradation]
            / (x[C.pErbB13_degradation] + x[C.init_RTKph] * x[C.pErbB13i_phosphatase])
            + x[C.Met_EGFR_basal_act]
            * x[C.init_EGFR]
            * x[C.init_Met]
            * x[C.pMetEGFR_degradation]
            / (x[C.pMetEGFR_degradation] + x[C.init_RTKph] * x[C.pMetEGFRi_phosphatase])
        )
        # 2 ∅ −→ ErbB2
        v[2] = (
            2
            * x[C.ErbB2_dimerize]
            * (x[C.init_ErbB2] ** 2)
            * x[C.pErbB2_degradation]
            / (x[C.pErbB2_degradation] + x[C.init_RTKph] * x[C.pErbB2i_phosphatase])
            + x[C.EGFR_ErbB2_basal_act]
            * x[C.init_EGFR]
            * x[C.init_ErbB2]
            * x[C.pErbB12_degradation]
            / (x[C.pErbB12_degradation] + x[C.init_RTKph] * x[C.pErbB12i_phosphatase])
            + x[C.ErbB3_ErbB2_basal_act]
            * x[C.init_ErbB2]
            * x[C.init_ErbB3]
            * x[C.pErbB32_degradation]
            / (x[C.pErbB32_degradation] + x[C.init_RTKph] * x[C.pErbB32i_phosphatase])
        )
        # 3 ∅ −→ ErbB3
        v[3] = (
            2
            * x[C.ErbB3_basal_activation]
            * (x[C.init_ErbB3] ** 2)
            * x[C.pErbB3_degradation]
            / (x[C.pErbB3_degradation] + x[C.init_RTKph] * x[C.pErbB3i_phosphatase])
            + x[C.EGFR_ErbB3_basal_act]
            * x[C.init_EGFR]
            * x[C.init_ErbB3]
            * x[C.pErbB13_degradation]
            / (x[C.pErbB13_degradation] + x[C.init_RTKph] * x[C.pErbB13i_phosphatase])
            + x[C.ErbB3_ErbB2_basal_act]
            * x[C.init_ErbB2]
            * x[C.init_ErbB3]
            * x[C.pErbB32_degradation]
            / (x[C.pErbB32_degradation] + x[C.init_RTKph] * x[C.pErbB32i_phosphatase])
            + x[C.Met_ErbB3_basal_act]
            * x[C.init_Met]
            * x[C.init_ErbB3]
            * x[C.pMetErbB3_degradation]
            / (x[C.pMetErbB3_degradation] + x[C.init_RTKph] * x[C.pMetErbB3i_phosphatase])
        )
        # 4 ∅ −→ IGF1R
        v[4] = (
            2
            * x[C.IGF1R_basal_activation]
            * (x[C.init_IGF1R] ** 2)
            * x[C.pIGF1R_degradation]
            / (x[C.pIGF1R_degradation] + x[C.init_RTKph] * x[C.pIGF1Ri_phosphatase])
        )
        # 5 ∅ −→ Met
        v[5] = (
            2
            * x[C.Met_basal_act]
            * (x[C.init_Met] ** 2)
            * x[C.pMet_degradation]
            / (x[C.pMet_degradation] + x[C.init_RTKph] * x[C.pMeti_phosphatase])
            + x[C.Met_EGFR_basal_act]
            * x[C.init_EGFR]
            * x[C.init_Met]
            * x[C.pMetEGFR_degradation]
            / (x[C.pMetEGFR_degradation] + x[C.init_RTKph] * x[C.pMetEGFRi_phosphatase])
            + x[C.Met_ErbB3_basal_act]
            * x[C.init_Met]
            * x[C.init_ErbB3]
            * x[C.pMetErbB3_degradation]
            / (x[C.pMetErbB3_degradation] + x[C.init_RTKph] * x[C.pMetErbB3i_phosphatase])
        )
        # 6 dose EGF + EGFR −→ EGFR EGF
        v[6] = x[C.EGFR_lig_binding] * y[V.dose_EGF] * y[V.EGFR]
        # 7 EGFR EGF −→ dose EGF + EGFR
        v[7] = x[C.EGF_kD] * y[V.EGFR_EGF] * x[C.EGFR_lig_binding]
        # test8 dose BTC + EGFR −→ EGFR BTC
        v[8] = y[V.EGFR] * x[C.EGFR_BTC_binding] * y[V.dose_BTC]
        # test9 EGFR BTC −→ dose BTC + EGFR
        v[9] = y[V.EGFR_BTC] * x[C.EGFR_BTC_binding] * x[C.EGF_kD]
        # 10 dose HRG + ErbB3 −→ ErbB3 HRG
        v[10] = x[C.ErbB3_lig_binding] * y[V.dose_HRG] * y[V.ErbB3]
        # 11 ErbB3 HRG −→ dose HRG + ErbB3
        v[11] = y[V.ErbB3_HRG] * x[C.ErbB3_lig_binding] * x[C.HRG_kD]
        # 12 dose IGF1 + IGF1R −→ IGF1R IGF1
        v[12] = x[C.IGF1R_lig_binding] * y[V.dose_IGF1] * y[V.IGF1R]
        # 13 IGF1R IGF1 −→ dose IGF1 + IGF1R
        v[13] = x[C.IGF1_kD] * y[V.IGF1R_IGF1] * x[C.IGF1R_lig_binding]
        # 14 dose HGF + Met −→ Met HGF
        v[14] = x[C.Met_lig_binding] * y[V.dose_HGF] * y[V.Met]
        # 15 Met HGF −→ dose HGF + Met
        v[15] = x[C.HGF_kD] * y[V.Met_HGF] * x[C.Met_lig_binding]
        # 16 2·EGFR EGF −→ pEGFRd
        v[16] = x[C.EGFR_dimerize] * (y[V.EGFR_EGF] ** 2)
        # test17 2·EGFR BTC −→ pEGFRd
        v[17] = (y[V.EGFR_BTC] ** 2) * x[C.EGFR_BTC_dimerize]
        # 18 2·ErbB2 −→ pErbB2
        v[18] = x[C.ErbB2_dimerize] * (y[V.ErbB2] ** 2)
        # 19 2·ErbB3 HRG −→ pErbB3d
        v[19] = x[C.ErbB3_dimerize] * (y[V.ErbB3_HRG] ** 2)
        # 20 2·IGF1R IGF1 −→ pIGF1Rd
        v[20] = x[C.IGF1R_dimerize] * (y[V.IGF1R_IGF1] ** 2)
        # 21 EGFR EGF + ErbB2 −→ pErbB12
        v[21] = x[C.EGFR_ErbB2_dimerize] * y[V.EGFR_EGF] * y[V.ErbB2]
        # test22 EGFR BTC + ErbB2 −→ pErbB12
        v[22] = y[V.EGFR_BTC] * x[C.EGFR_ErbB2_BTC_dimerize] * y[V.ErbB2]
        # 23 EGFR EGF + ErbB3 HRG −→ pErbB13
        v[23] = x[C.EGFR_ErbB3_dimerize] * y[V.EGFR_EGF] * y[V.ErbB3_HRG]
        # test24 EGFR BTC + ErbB3 HRG −→ pErbB13
        v[24] = y[V.EGFR_BTC] * x[C.EGFR_ErbB3_BTC_dimerize] * y[V.ErbB3_HRG]
        # test25 EGFR BTC + ErbB3 −→ pErbB13
        v[25] = y[V.EGFR_BTC] * x[C.EGFR_ErbB3_dimerize_noHRG] * y[V.ErbB3]
        # 26 ErbB2 + ErbB3 HRG −→ pErbB32
        v[26] = x[C.ErbB2_ErbB3_dimerize] * y[V.ErbB2] * y[V.ErbB3_HRG]
        # 27 2·Met HGF −→ pMetd
        v[27] = x[C.Met_dimerize] * (y[V.Met_HGF] ** 2)
        # 28 ErbB3 HRG + Met −→ pMetErbB3
        v[28] = x[C.Met_ErbB3_dimerize] * y[V.ErbB3_HRG] * y[V.Met]
        # 29 ErbB3 HRG + Met HGF −→ pMetErbB3
        v[29] = x[C.Met_lig_ErbB3_dimerize] * y[V.ErbB3_HRG] * y[V.Met_HGF]
        # 30 EGFR EGF + Met HGF −→ pMetEGFR
        v[30] = x[C.Met_EGFR_dimerize] * y[V.EGFR_EGF] * y[V.Met_HGF]
        # test31 EGFR BTC + Met HGF −→ pMetEGFR
        v[31] = y[V.EGFR_BTC] * x[C.Met_EGFR_BTC_dimerize] * y[V.Met_HGF]
        # 32 2·EGFR −→ pEGFRd
        v[32] = x[C.EGFR_basal_activation] * (y[V.EGFR] ** 2)
        # 33 2·ErbB3 −→ pErbB3d
        v[33] = x[C.ErbB3_basal_activation] * (y[V.ErbB3] ** 2)
        # 34 2·IGF1R −→ pIGF1Rd
        v[34] = x[C.IGF1R_basal_activation] * (y[V.IGF1R] ** 2)
        # 35 EGFR + ErbB2 −→ pErbB12
        v[35] = x[C.EGFR_ErbB2_basal_act] * y[V.EGFR] * y[V.ErbB2]
        # 36 EGFR + ErbB3 −→ pErbB13
        v[36] = x[C.EGFR_ErbB3_basal_act] * y[V.EGFR] * y[V.ErbB3]
        # 37 ErbB2 + ErbB3 −→ pErbB32
        v[37] = x[C.ErbB3_ErbB2_basal_act] * y[V.ErbB2] * y[V.ErbB3]
        # 38 ErbB3 + Met −→ pMetErbB3
        v[38] = x[C.Met_ErbB3_basal_act] * y[V.ErbB3] * y[V.Met]
        # 39 2·Met −→ pMetd
        v[39] = x[C.Met_basal_act] * (y[V.Met] ** 2)
        # 40 EGFR + Met −→ pMetEGFR
        v[40] = x[C.Met_EGFR_basal_act] * y[V.EGFR] * y[V.Met]
        # 41 pEGFRd −→ pEGFRi
        v[41] = x[C.pEGFR_internalize] * y[V.pEGFRd]
        # 42 pEGFRi −→ ∅
        v[42] = x[C.pEGFR_degradation] * y[V.pEGFRi]
        # 43 RTKph + pEGFRi −→ pEGFRi ph
        v[43] = x[C.pEGFR_phosphatase_binding] * y[V.RTKph] * y[V.pEGFRi]
        # 44 pEGFRi ph −→ RTKph + 2·EGFRi
        v[44] = x[C.pEGFRi_dephosph] * y[V.pEGFRi_ph]
        # 45 EGFRi −→ EGFR
        v[45] = x[C.EGFR_basal_recycle] * y[V.EGFRi]
        # 46 pErbB2 −→ pErbB2i
        v[46] = x[C.pErbB2_internalize] * y[V.pErbB2]
        # 47 RTKph + pErbB2i −→ pErbB2i ph
        v[47] = x[C.pErbB2i_phosphatase] * y[V.RTKph] * y[V.pErbB2i]
        # 48 pErbB2i ph −→ RTKph + 2·ErbB2i
        v[48] = x[C.pErbB2i_dephosph] * y[V.pErbB2i_ph]
        # 49 pErbB2i −→ ∅
        v[49] = x[C.pErbB2_degradation] * y[V.pErbB2i]
        # 50 ErbB2i −→ ErbB2
        v[50] = x[C.ErbB2_recycle] * y[V.ErbB2i]
        # 51 pErbB3d −→ pErbB3i
        v[51] = x[C.pErbB3_internalize] * y[V.pErbB3d]
        # 52 pErbB3i −→ ∅
        v[52] = x[C.pErbB3_degradation] * y[V.pErbB3i]
        # 53 RTKph + pErbB3i −→ pErbB3i ph
        v[53] = x[C.pErbB3i_phosphatase] * y[V.RTKph] * y[V.pErbB3i]
        # 54 pErbB3i ph −→ RTKph + 2·ErbB3i
        v[54] = x[C.pErbB3i_dephosph] * y[V.pErbB3i_ph]
        # 55 ErbB3i −→ ErbB3
        v[55] = x[C.ErbB3_basal_recycle] * y[V.ErbB3i]
        # 56 pIGF1Rd −→ pIGF1Ri
        v[56] = x[C.pIGF1R_internalize] * y[V.pIGF1Rd]
        # 57 pIGF1Ri −→ ∅
        v[57] = x[C.pIGF1R_degradation] * y[V.pIGF1Ri]
        # 58 RTKph + pIGF1Ri −→ pIGF1Ri ph
        v[58] = x[C.pIGF1Ri_phosphatase] * y[V.RTKph] * y[V.pIGF1Ri]
        # 59 pIGF1Ri ph −→ RTKph + 2·IGF1Ri
        v[59] = x[C.pIGF1Ri_dephosph] * y[V.pIGF1Ri_ph]
        # 60 IGF1Ri −→ IGF1R
        v[60] = x[C.IGF1R_basal_recycle] * y[V.IGF1Ri]
        # 61 pErbB12 −→ pErbB12i
        v[61] = x[C.pErbB12_internalize] * y[V.pErbB12]
        # 62 pErbB12i −→ ∅
        v[62] = x[C.pErbB12_degradation] * y[V.pErbB12i]
        # 63 RTKph + pErbB12i −→ pErbB12i ph
        v[63] = x[C.pErbB12i_phosphatase] * y[V.RTKph] * y[V.pErbB12i]
        # 64 pErbB12i ph −→ RTKph + EGFRi + ErbB2i
        v[64] = x[C.pErbB12i_dephosph] * y[V.pErbB12i_ph]
        # 65 pErbB32 −→ pErbB32i
        v[65] = x[C.pErbB32_internalize] * y[V.pErbB32]
        # 66 pErbB32i −→ ∅
        v[66] = x[C.pErbB32_degradation] * y[V.pErbB32i]
        # 67 RTKph + pErbB32i −→ pErbB32i ph
        v[67] = x[C.pErbB32i_phosphatase] * y[V.RTKph] * y[V.pErbB32i]
        # 68 pErbB32i ph −→ RTKph + ErbB2i + ErbB3i
        v[68] = x[C.pErbB32i_dephosph] * y[V.pErbB32i_ph]
        # 69 pErbB13 −→ pErbB13i
        v[69] = x[C.pErbB13_internalize] * y[V.pErbB13]
        # 70 pErbB13i −→ ∅
        v[70] = x[C.pErbB13_degradation] * y[V.pErbB13i]
        # 71 RTKph + pErbB13i −→ pErbB13i ph
        v[71] = x[C.pErbB13i_phosphatase] * y[V.RTKph] * y[V.pErbB13i]
        # 72 pErbB13i ph −→ RTKph + EGFRi + ErbB3i
        v[72] = x[C.pErbB13i_dephosph] * y[V.pErbB13i_ph]
        # 73 pMetd −→ pMeti
        v[73] = x[C.pMet_internalize] * y[V.pMetd]
        # 74 pMeti −→ ∅
        v[74] = x[C.pMet_degradation] * y[V.pMeti]
        # 75 RTKph + pMeti −→ pMeti ph
        v[75] = x[C.pMeti_phosphatase] * y[V.RTKph] * y[V.pMeti]
        # 76 pMeti ph −→ RTKph + 2·Meti
        v[76] = x[C.pMeti_dephosph] * y[V.pMeti_ph]
        # 77 Meti −→ Met
        v[77] = x[C.Met_recycle] * y[V.Meti]
        # 78 pMetErbB3 −→ pMetErbB3i
        v[78] = x[C.pMetErbB3_internalize] * y[V.pMetErbB3]
        # 79 pMetErbB3i −→ ∅
        v[79] = x[C.pMetErbB3_degradation] * y[V.pMetErbB3i]
        # 80 RTKph + pMetErbB3i −→ pMetErbB3i ph
        v[80] = x[C.pMetErbB3i_phosphatase] * y[V.RTKph] * y[V.pMetErbB3i]
        # 81 pMetErbB3i ph −→ RTKph + ErbB3i + Meti
        v[81] = x[C.pMetErbB3i_dephosph] * y[V.pMetErbB3i_ph]
        # 82 pMetEGFR −→ pMetEGFRi
        v[82] = x[C.pMetEGFR_internalize] * y[V.pMetEGFR]
        # 83 pMetEGFRi −→ ∅
        v[83] = x[C.pMetEGFR_degradation] * y[V.pMetEGFRi]
        # 84 RTKph + pMetEGFRi −→ pMetEGFRi ph
        v[84] = x[C.pMetEGFRi_phosphatase] * y[V.RTKph] * y[V.pMetEGFRi]
        # 85 pMetEGFRi ph −→ RTKph + EGFRi + Meti
        v[85] = x[C.pMetEGFRi_dephosph] * y[V.pMetEGFRi_ph]
        # 86 MEK -pEGFRd,pERK→ pMEK
        v[86] = (
            y[V.MEK]
            * x[C.MEK_phosphorylation_pEGFR]
            * y[V.pEGFRd]
            / (
                x[C.feedback_pERK] * y[V.pERK]
                + x[C.feedback_pAKT]
                * x[C.init_AKT]
                * (
                    x[C.AKT_activation_pEGFR]
                    * x[C.EGFR_basal_activation]
                    * (x[C.init_EGFR] ** 2)
                    / x[C.pEGFR_internalize]
                    + x[C.AKT_activation_pIGF1R]
                    * x[C.IGF1R_basal_activation]
                    * (x[C.init_IGF1R] ** 2)
                    / x[C.pIGF1R_internalize]
                    + x[C.AKT_activation_pMetd]
                    * x[C.Met_basal_act]
                    * (x[C.init_Met] ** 2)
                    / x[C.pMet_internalize]
                    + x[C.AKT_activation_pIGF1R]
                    * x[C.AKT_internIGF1R_effect]
                    * x[C.IGF1R_basal_activation]
                    * (x[C.init_IGF1R] ** 2)
                    / (x[C.pIGF1R_degradation] + x[C.init_RTKph] * x[C.pIGF1Ri_phosphatase])
                    + x[C.AKT_activation_pErbB12]
                    * x[C.EGFR_ErbB2_basal_act]
                    * x[C.init_EGFR]
                    * x[C.init_ErbB2]
                    / x[C.pErbB12_internalize]
                    + x[C.AKT_activation_pErbB13]
                    * x[C.EGFR_ErbB3_basal_act]
                    * x[C.init_EGFR]
                    * x[C.init_ErbB3]
                    / x[C.pErbB13_internalize]
                    + x[C.AKT_activation_pErbB32]
                    * x[C.ErbB3_ErbB2_basal_act]
                    * x[C.init_ErbB2]
                    * x[C.init_ErbB3]
                    / x[C.pErbB32_internalize]
                    + x[C.AKT_activation_pMetEGFR]
                    * x[C.Met_EGFR_basal_act]
                    * x[C.init_EGFR]
                    * x[C.init_Met]
                    / x[C.pMetEGFR_internalize]
                    + x[C.AKT_activation_pMetErbB3]
                    * x[C.Met_ErbB3_basal_act]
                    * x[C.init_Met]
                    * x[C.init_ErbB3]
                    / x[C.pMetErbB3_internalize]
                )
                / (
                    x[C.pAKT_deactivation]
                    * (
                        x[C.feedback_pS6K1] * x[C.init_pS6K1]
                        + x[C.feedback_pERK_on_AKT] * x[C.init_pERK]
                        + 1
                    )
                )
                + 1
            )
        )
        # 87 MEK -pErbB12,pERK→  pMEK
        v[87] = (
            y[V.MEK]
            * x[C.MEK_phosphorylation_pErbB12]
            * y[V.pErbB12]
            / (
                x[C.feedback_pERK] * y[V.pERK]
                + x[C.feedback_pAKT]
                * x[C.init_AKT]
                * (
                    x[C.AKT_activation_pEGFR]
                    * x[C.EGFR_basal_activation]
                    * (x[C.init_EGFR] ** 2)
                    / x[C.pEGFR_internalize]
                    + x[C.AKT_activation_pIGF1R]
                    * x[C.IGF1R_basal_activation]
                    * (x[C.init_IGF1R] ** 2)
                    / x[C.pIGF1R_internalize]
                    + x[C.AKT_activation_pMetd]
                    * x[C.Met_basal_act]
                    * (x[C.init_Met] ** 2)
                    / x[C.pMet_internalize]
                    + x[C.AKT_activation_pIGF1R]
                    * x[C.AKT_internIGF1R_effect]
                    * x[C.IGF1R_basal_activation]
                    * (x[C.init_IGF1R] ** 2)
                    / (x[C.pIGF1R_degradation] + x[C.init_RTKph] * x[C.pIGF1Ri_phosphatase])
                    + x[C.AKT_activation_pErbB12]
                    * x[C.EGFR_ErbB2_basal_act]
                    * x[C.init_EGFR]
                    * x[C.init_ErbB2]
                    / x[C.pErbB12_internalize]
                    + x[C.AKT_activation_pErbB13]
                    * x[C.EGFR_ErbB3_basal_act]
                    * x[C.init_EGFR]
                    * x[C.init_ErbB3]
                    / x[C.pErbB13_internalize]
                    + x[C.AKT_activation_pErbB32]
                    * x[C.ErbB3_ErbB2_basal_act]
                    * x[C.init_ErbB2]
                    * x[C.init_ErbB3]
                    / x[C.pErbB32_internalize]
                    + x[C.AKT_activation_pMetEGFR]
                    * x[C.Met_EGFR_basal_act]
                    * x[C.init_EGFR]
                    * x[C.init_Met]
                    / x[C.pMetEGFR_internalize]
                    + x[C.AKT_activation_pMetErbB3]
                    * x[C.Met_ErbB3_basal_act]
                    * x[C.init_Met]
                    * x[C.init_ErbB3]
                    / x[C.pMetErbB3_internalize]
                )
                / (
                    x[C.pAKT_deactivation]
                    * (
                        x[C.feedback_pS6K1] * x[C.init_pS6K1]
                        + x[C.feedback_pERK_on_AKT] * x[C.init_pERK]
                        + 1
                    )
                )
                + 1
            )
        )
        # 88 MEK -pErbB13,pERK→  pMEK
        v[88] = (
            y[V.MEK]
            * x[C.MEK_phosphorylation_pErbB13]
            * y[V.pErbB13]
            / (
                x[C.feedback_pERK] * y[V.pERK]
                + x[C.feedback_pAKT]
                * x[C.init_AKT]
                * (
                    x[C.AKT_activation_pEGFR]
                    * x[C.EGFR_basal_activation]
                    * (x[C.init_EGFR] ** 2)
                    / x[C.pEGFR_internalize]
                    + x[C.AKT_activation_pIGF1R]
                    * x[C.IGF1R_basal_activation]
                    * (x[C.init_IGF1R] ** 2)
                    / x[C.pIGF1R_internalize]
                    + x[C.AKT_activation_pMetd]
                    * x[C.Met_basal_act]
                    * (x[C.init_Met] ** 2)
                    / x[C.pMet_internalize]
                    + x[C.AKT_activation_pIGF1R]
                    * x[C.AKT_internIGF1R_effect]
                    * x[C.IGF1R_basal_activation]
                    * (x[C.init_IGF1R] ** 2)
                    / (x[C.pIGF1R_degradation] + x[C.init_RTKph] * x[C.pIGF1Ri_phosphatase])
                    + x[C.AKT_activation_pErbB12]
                    * x[C.EGFR_ErbB2_basal_act]
                    * x[C.init_EGFR]
                    * x[C.init_ErbB2]
                    / x[C.pErbB12_internalize]
                    + x[C.AKT_activation_pErbB13]
                    * x[C.EGFR_ErbB3_basal_act]
                    * x[C.init_EGFR]
                    * x[C.init_ErbB3]
                    / x[C.pErbB13_internalize]
                    + x[C.AKT_activation_pErbB32]
                    * x[C.ErbB3_ErbB2_basal_act]
                    * x[C.init_ErbB2]
                    * x[C.init_ErbB3]
                    / x[C.pErbB32_internalize]
                    + x[C.AKT_activation_pMetEGFR]
                    * x[C.Met_EGFR_basal_act]
                    * x[C.init_EGFR]
                    * x[C.init_Met]
                    / x[C.pMetEGFR_internalize]
                    + x[C.AKT_activation_pMetErbB3]
                    * x[C.Met_ErbB3_basal_act]
                    * x[C.init_Met]
                    * x[C.init_ErbB3]
                    / x[C.pMetErbB3_internalize]
                )
                / (
                    x[C.pAKT_deactivation]
                    * (
                        x[C.feedback_pS6K1] * x[C.init_pS6K1]
                        + x[C.feedback_pERK_on_AKT] * x[C.init_pERK]
                        + 1
                    )
                )
                + 1
            )
        )
        # 89 MEK -pErbB32,pERK→ pMEK
        v[89] = (
            y[V.MEK]
            * x[C.MEK_phosphorylation_pErbB32]
            * y[V.pErbB32]
            / (
                x[C.feedback_pERK] * y[V.pERK]
                + x[C.feedback_pAKT]
                * x[C.init_AKT]
                * (
                    x[C.AKT_activation_pEGFR]
                    * x[C.EGFR_basal_activation]
                    * (x[C.init_EGFR] ** 2)
                    / x[C.pEGFR_internalize]
                    + x[C.AKT_activation_pIGF1R]
                    * x[C.IGF1R_basal_activation]
                    * (x[C.init_IGF1R] ** 2)
                    / x[C.pIGF1R_internalize]
                    + x[C.AKT_activation_pMetd]
                    * x[C.Met_basal_act]
                    * (x[C.init_Met] ** 2)
                    / x[C.pMet_internalize]
                    + x[C.AKT_activation_pIGF1R]
                    * x[C.AKT_internIGF1R_effect]
                    * x[C.IGF1R_basal_activation]
                    * (x[C.init_IGF1R] ** 2)
                    / (x[C.pIGF1R_degradation] + x[C.init_RTKph] * x[C.pIGF1Ri_phosphatase])
                    + x[C.AKT_activation_pErbB12]
                    * x[C.EGFR_ErbB2_basal_act]
                    * x[C.init_EGFR]
                    * x[C.init_ErbB2]
                    / x[C.pErbB12_internalize]
                    + x[C.AKT_activation_pErbB13]
                    * x[C.EGFR_ErbB3_basal_act]
                    * x[C.init_EGFR]
                    * x[C.init_ErbB3]
                    / x[C.pErbB13_internalize]
                    + x[C.AKT_activation_pErbB32]
                    * x[C.ErbB3_ErbB2_basal_act]
                    * x[C.init_ErbB2]
                    * x[C.init_ErbB3]
                    / x[C.pErbB32_internalize]
                    + x[C.AKT_activation_pMetEGFR]
                    * x[C.Met_EGFR_basal_act]
                    * x[C.init_EGFR]
                    * x[C.init_Met]
                    / x[C.pMetEGFR_internalize]
                    + x[C.AKT_activation_pMetErbB3]
                    * x[C.Met_ErbB3_basal_act]
                    * x[C.init_Met]
                    * x[C.init_ErbB3]
                    / x[C.pMetErbB3_internalize]
                )
                / (
                    x[C.pAKT_deactivation]
                    * (
                        x[C.feedback_pS6K1] * x[C.init_pS6K1]
                        + x[C.feedback_pERK_on_AKT] * x[C.init_pERK]
                        + 1
                    )
                )
                + 1
            )
        )
        # 90 MEK -pMetd,pERK→ pMEK
        v[90] = (
            y[V.MEK]
            * x[C.MEK_phosphorylation_pMetd]
            * y[V.pMetd]
            / (
                x[C.feedback_pERK] * y[V.pERK]
                + x[C.feedback_pAKT]
                * x[C.init_AKT]
                * (
                    x[C.AKT_activation_pEGFR]
                    * x[C.EGFR_basal_activation]
                    * (x[C.init_EGFR] ** 2)
                    / x[C.pEGFR_internalize]
                    + x[C.AKT_activation_pIGF1R]
                    * x[C.IGF1R_basal_activation]
                    * (x[C.init_IGF1R] ** 2)
                    / x[C.pIGF1R_internalize]
                    + x[C.AKT_activation_pMetd]
                    * x[C.Met_basal_act]
                    * (x[C.init_Met] ** 2)
                    / x[C.pMet_internalize]
                    + x[C.AKT_activation_pIGF1R]
                    * x[C.AKT_internIGF1R_effect]
                    * x[C.IGF1R_basal_activation]
                    * (x[C.init_IGF1R] ** 2)
                    / (x[C.pIGF1R_degradation] + x[C.init_RTKph] * x[C.pIGF1Ri_phosphatase])
                    + x[C.AKT_activation_pErbB12]
                    * x[C.EGFR_ErbB2_basal_act]
                    * x[C.init_EGFR]
                    * x[C.init_ErbB2]
                    / x[C.pErbB12_internalize]
                    + x[C.AKT_activation_pErbB13]
                    * x[C.EGFR_ErbB3_basal_act]
                    * x[C.init_EGFR]
                    * x[C.init_ErbB3]
                    / x[C.pErbB13_internalize]
                    + x[C.AKT_activation_pErbB32]
                    * x[C.ErbB3_ErbB2_basal_act]
                    * x[C.init_ErbB2]
                    * x[C.init_ErbB3]
                    / x[C.pErbB32_internalize]
                    + x[C.AKT_activation_pMetEGFR]
                    * x[C.Met_EGFR_basal_act]
                    * x[C.init_EGFR]
                    * x[C.init_Met]
                    / x[C.pMetEGFR_internalize]
                    + x[C.AKT_activation_pMetErbB3]
                    * x[C.Met_ErbB3_basal_act]
                    * x[C.init_Met]
                    * x[C.init_ErbB3]
                    / x[C.pMetErbB3_internalize]
                )
                / (
                    x[C.pAKT_deactivation]
                    * (
                        x[C.feedback_pS6K1] * x[C.init_pS6K1]
                        + x[C.feedback_pERK_on_AKT] * x[C.init_pERK]
                        + 1
                    )
                )
                + 1
            )
        )
        # 91 MEK -pMetEGFR,pERK→ pMEK
        v[91] = (
            y[V.MEK]
            * x[C.MEK_phosphorylation_pMetEGFR]
            * y[V.pMetEGFR]
            / (
                x[C.feedback_pERK] * y[V.pERK]
                + x[C.feedback_pAKT]
                * x[C.init_AKT]
                * (
                    x[C.AKT_activation_pEGFR]
                    * x[C.EGFR_basal_activation]
                    * (x[C.init_EGFR] ** 2)
                    / x[C.pEGFR_internalize]
                    + x[C.AKT_activation_pIGF1R]
                    * x[C.IGF1R_basal_activation]
                    * (x[C.init_IGF1R] ** 2)
                    / x[C.pIGF1R_internalize]
                    + x[C.AKT_activation_pMetd]
                    * x[C.Met_basal_act]
                    * (x[C.init_Met] ** 2)
                    / x[C.pMet_internalize]
                    + x[C.AKT_activation_pIGF1R]
                    * x[C.AKT_internIGF1R_effect]
                    * x[C.IGF1R_basal_activation]
                    * (x[C.init_IGF1R] ** 2)
                    / (x[C.pIGF1R_degradation] + x[C.init_RTKph] * x[C.pIGF1Ri_phosphatase])
                    + x[C.AKT_activation_pErbB12]
                    * x[C.EGFR_ErbB2_basal_act]
                    * x[C.init_EGFR]
                    * x[C.init_ErbB2]
                    / x[C.pErbB12_internalize]
                    + x[C.AKT_activation_pErbB13]
                    * x[C.EGFR_ErbB3_basal_act]
                    * x[C.init_EGFR]
                    * x[C.init_ErbB3]
                    / x[C.pErbB13_internalize]
                    + x[C.AKT_activation_pErbB32]
                    * x[C.ErbB3_ErbB2_basal_act]
                    * x[C.init_ErbB2]
                    * x[C.init_ErbB3]
                    / x[C.pErbB32_internalize]
                    + x[C.AKT_activation_pMetEGFR]
                    * x[C.Met_EGFR_basal_act]
                    * x[C.init_EGFR]
                    * x[C.init_Met]
                    / x[C.pMetEGFR_internalize]
                    + x[C.AKT_activation_pMetErbB3]
                    * x[C.Met_ErbB3_basal_act]
                    * x[C.init_Met]
                    * x[C.init_ErbB3]
                    / x[C.pMetErbB3_internalize]
                )
                / (
                    x[C.pAKT_deactivation]
                    * (
                        x[C.feedback_pS6K1] * x[C.init_pS6K1]
                        + x[C.feedback_pERK_on_AKT] * x[C.init_pERK]
                        + 1
                    )
                )
                + 1
            )
        )
        # 92 MEK -pIGF1Rd,pERK→ pMEK
        v[92] = (
            y[V.MEK]
            * x[C.MEK_phosphorylation_pIGF1R]
            * y[V.pIGF1Rd]
            / (
                x[C.feedback_pERK] * y[V.pERK]
                + x[C.feedback_pAKT]
                * x[C.init_AKT]
                * (
                    x[C.AKT_activation_pEGFR]
                    * x[C.EGFR_basal_activation]
                    * (x[C.init_EGFR] ** 2)
                    / x[C.pEGFR_internalize]
                    + x[C.AKT_activation_pIGF1R]
                    * x[C.IGF1R_basal_activation]
                    * (x[C.init_IGF1R] ** 2)
                    / x[C.pIGF1R_internalize]
                    + x[C.AKT_activation_pMetd]
                    * x[C.Met_basal_act]
                    * (x[C.init_Met] ** 2)
                    / x[C.pMet_internalize]
                    + x[C.AKT_activation_pIGF1R]
                    * x[C.AKT_internIGF1R_effect]
                    * x[C.IGF1R_basal_activation]
                    * (x[C.init_IGF1R] ** 2)
                    / (x[C.pIGF1R_degradation] + x[C.init_RTKph] * x[C.pIGF1Ri_phosphatase])
                    + x[C.AKT_activation_pErbB12]
                    * x[C.EGFR_ErbB2_basal_act]
                    * x[C.init_EGFR]
                    * x[C.init_ErbB2]
                    / x[C.pErbB12_internalize]
                    + x[C.AKT_activation_pErbB13]
                    * x[C.EGFR_ErbB3_basal_act]
                    * x[C.init_EGFR]
                    * x[C.init_ErbB3]
                    / x[C.pErbB13_internalize]
                    + x[C.AKT_activation_pErbB32]
                    * x[C.ErbB3_ErbB2_basal_act]
                    * x[C.init_ErbB2]
                    * x[C.init_ErbB3]
                    / x[C.pErbB32_internalize]
                    + x[C.AKT_activation_pMetEGFR]
                    * x[C.Met_EGFR_basal_act]
                    * x[C.init_EGFR]
                    * x[C.init_Met]
                    / x[C.pMetEGFR_internalize]
                    + x[C.AKT_activation_pMetErbB3]
                    * x[C.Met_ErbB3_basal_act]
                    * x[C.init_Met]
                    * x[C.init_ErbB3]
                    / x[C.pMetErbB3_internalize]
                )
                / (
                    x[C.pAKT_deactivation]
                    * (
                        x[C.feedback_pS6K1] * x[C.init_pS6K1]
                        + x[C.feedback_pERK_on_AKT] * x[C.init_pERK]
                        + 1
                    )
                )
                + 1
            )
        )
        # 93 MEK -pMetErbB3,pERK→ pMEK
        v[93] = (
            y[V.MEK]
            * x[C.MEK_phosphorylation_pMetErbB3]
            * y[V.pMetErbB3]
            / (
                x[C.feedback_pERK] * y[V.pERK]
                + x[C.feedback_pAKT]
                * x[C.init_AKT]
                * (
                    x[C.AKT_activation_pEGFR]
                    * x[C.EGFR_basal_activation]
                    * (x[C.init_EGFR] ** 2)
                    / x[C.pEGFR_internalize]
                    + x[C.AKT_activation_pIGF1R]
                    * x[C.IGF1R_basal_activation]
                    * (x[C.init_IGF1R] ** 2)
                    / x[C.pIGF1R_internalize]
                    + x[C.AKT_activation_pMetd]
                    * x[C.Met_basal_act]
                    * (x[C.init_Met] ** 2)
                    / x[C.pMet_internalize]
                    + x[C.AKT_activation_pIGF1R]
                    * x[C.AKT_internIGF1R_effect]
                    * x[C.IGF1R_basal_activation]
                    * (x[C.init_IGF1R] ** 2)
                    / (x[C.pIGF1R_degradation] + x[C.init_RTKph] * x[C.pIGF1Ri_phosphatase])
                    + x[C.AKT_activation_pErbB12]
                    * x[C.EGFR_ErbB2_basal_act]
                    * x[C.init_EGFR]
                    * x[C.init_ErbB2]
                    / x[C.pErbB12_internalize]
                    + x[C.AKT_activation_pErbB13]
                    * x[C.EGFR_ErbB3_basal_act]
                    * x[C.init_EGFR]
                    * x[C.init_ErbB3]
                    / x[C.pErbB13_internalize]
                    + x[C.AKT_activation_pErbB32]
                    * x[C.ErbB3_ErbB2_basal_act]
                    * x[C.init_ErbB2]
                    * x[C.init_ErbB3]
                    / x[C.pErbB32_internalize]
                    + x[C.AKT_activation_pMetEGFR]
                    * x[C.Met_EGFR_basal_act]
                    * x[C.init_EGFR]
                    * x[C.init_Met]
                    / x[C.pMetEGFR_internalize]
                    + x[C.AKT_activation_pMetErbB3]
                    * x[C.Met_ErbB3_basal_act]
                    * x[C.init_Met]
                    * x[C.init_ErbB3]
                    / x[C.pMetErbB3_internalize]
                )
                / (
                    x[C.pAKT_deactivation]
                    * (
                        x[C.feedback_pS6K1] * x[C.init_pS6K1]
                        + x[C.feedback_pERK_on_AKT] * x[C.init_pERK]
                        + 1
                    )
                )
                + 1
            )
        )
        # 94 MEK -pIGF1Ri,pERK→ pMEK
        v[94] = (
            y[V.MEK]
            * x[C.MEK_internIGF1R_effect]
            * x[C.MEK_phosphorylation_pIGF1R]
            * y[V.pIGF1Ri]
            / (
                x[C.feedback_pERK] * y[V.pERK]
                + x[C.feedback_pAKT]
                * x[C.init_AKT]
                * (
                    x[C.AKT_activation_pEGFR]
                    * x[C.EGFR_basal_activation]
                    * (x[C.init_EGFR] ** 2)
                    / x[C.pEGFR_internalize]
                    + x[C.AKT_activation_pIGF1R]
                    * x[C.IGF1R_basal_activation]
                    * (x[C.init_IGF1R] ** 2)
                    / x[C.pIGF1R_internalize]
                    + x[C.AKT_activation_pMetd]
                    * x[C.Met_basal_act]
                    * (x[C.init_Met] ** 2)
                    / x[C.pMet_internalize]
                    + x[C.AKT_activation_pIGF1R]
                    * x[C.AKT_internIGF1R_effect]
                    * x[C.IGF1R_basal_activation]
                    * (x[C.init_IGF1R] ** 2)
                    / (x[C.pIGF1R_degradation] + x[C.init_RTKph] * x[C.pIGF1Ri_phosphatase])
                    + x[C.AKT_activation_pErbB12]
                    * x[C.EGFR_ErbB2_basal_act]
                    * x[C.init_EGFR]
                    * x[C.init_ErbB2]
                    / x[C.pErbB12_internalize]
                    + x[C.AKT_activation_pErbB13]
                    * x[C.EGFR_ErbB3_basal_act]
                    * x[C.init_EGFR]
                    * x[C.init_ErbB3]
                    / x[C.pErbB13_internalize]
                    + x[C.AKT_activation_pErbB32]
                    * x[C.ErbB3_ErbB2_basal_act]
                    * x[C.init_ErbB2]
                    * x[C.init_ErbB3]
                    / x[C.pErbB32_internalize]
                    + x[C.AKT_activation_pMetEGFR]
                    * x[C.Met_EGFR_basal_act]
                    * x[C.init_EGFR]
                    * x[C.init_Met]
                    / x[C.pMetEGFR_internalize]
                    + x[C.AKT_activation_pMetErbB3]
                    * x[C.Met_ErbB3_basal_act]
                    * x[C.init_Met]
                    * x[C.init_ErbB3]
                    / x[C.pMetErbB3_internalize]
                )
                / (
                    x[C.pAKT_deactivation]
                    * (
                        x[C.feedback_pS6K1] * x[C.init_pS6K1]
                        + x[C.feedback_pERK_on_AKT] * x[C.init_pERK]
                        + 1
                    )
                )
                + 1
            )
        )
        # 95 pMEK −→ MEK
        v[95] = x[C.pMEK_dephosphorylation] * y[V.pMEK]
        # 96 ERK -pMEK→ pERK
        v[96] = y[V.ERK] * x[C.ERK_phosphorylation_pMEK] * y[V.pMEK]
        # 97 pERK −→ ERK
        v[97] = x[C.pERK_dephosphorylation] * y[V.pERK]
        # 98 AKT -pEGFRd,pERK,pS6K1→ pAKT
        v[98] = (
            y[V.AKT]
            * x[C.AKT_activation_pEGFR]
            * y[V.pEGFRd]
            / (x[C.feedback_pERK_on_AKT] * y[V.pERK] + x[C.feedback_pS6K1] * y[V.pS6K1] + 1)
        )
        # 99 AKT -pErbB12,pERK,pS6K1→ pAKT
        v[99] = (
            y[V.AKT]
            * x[C.AKT_activation_pErbB12]
            * y[V.pErbB12]
            / (x[C.feedback_pERK_on_AKT] * y[V.pERK] + x[C.feedback_pS6K1] * y[V.pS6K1] + 1)
        )
        # 100 AKT -pErbB13,pERK,pS6K1→ pAKT
        v[100] = (
            y[V.AKT]
            * x[C.AKT_activation_pErbB13]
            * y[V.pErbB13]
            / (x[C.feedback_pERK_on_AKT] * y[V.pERK] + x[C.feedback_pS6K1] * y[V.pS6K1] + 1)
        )
        # 101 AKT -pErbB32, pERK,pS6K1→ pAKT
        v[101] = (
            y[V.AKT]
            * x[C.AKT_activation_pErbB32]
            * y[V.pErbB32]
            / (x[C.feedback_pERK_on_AKT] * y[V.pERK] + x[C.feedback_pS6K1] * y[V.pS6K1] + 1)
        )
        # 102 AKT -pMetEGFR,pERK,pS6K1→ pAKT
        v[102] = (
            y[V.AKT]
            * x[C.AKT_activation_pMetEGFR]
            * y[V.pMetEGFR]
            / (x[C.feedback_pERK_on_AKT] * y[V.pERK] + x[C.feedback_pS6K1] * y[V.pS6K1] + 1)
        )
        # 103 AKT -pMetd,pERK,pS6K1→ pAKT
        v[103] = (
            y[V.AKT]
            * x[C.AKT_activation_pMetd]
            * y[V.pMetd]
            / (x[C.feedback_pERK_on_AKT] * y[V.pERK] + x[C.feedback_pS6K1] * y[V.pS6K1] + 1)
        )
        # 104 AKT -pIGF1Rd,pERK,pS6K1→ pAKT
        v[104] = (
            y[V.AKT]
            * x[C.AKT_activation_pIGF1R]
            * y[V.pIGF1Rd]
            / (x[C.feedback_pERK_on_AKT] * y[V.pERK] + x[C.feedback_pS6K1] * y[V.pS6K1] + 1)
        )
        # 105 AKT -pMetErbB3,pERK,pS6K1→ pAKT
        v[105] = (
            y[V.AKT]
            * x[C.AKT_activation_pMetErbB3]
            * y[V.pMetErbB3]
            / (x[C.feedback_pERK_on_AKT] * y[V.pERK] + x[C.feedback_pS6K1] * y[V.pS6K1] + 1)
        )
        # 106 AKT −pIGF1Ri,pERK,pS6K1→ pAKT
        v[106] = (
            y[V.AKT]
            * x[C.AKT_activation_pIGF1R]
            * x[C.AKT_internIGF1R_effect]
            * y[V.pIGF1Ri]
            / (x[C.feedback_pERK_on_AKT] * y[V.pERK] + x[C.feedback_pS6K1] * y[V.pS6K1] + 1)
        )
        # 107 pAKT −→ AKT
        v[107] = x[C.pAKT_deactivation] * y[V.pAKT]
        # 108 -S6K1-pAKT→ pS6K1
        v[108] = y[V.S6K1] * x[C.S6K1_phosphorylation_pAKT] * y[V.pAKT]
        # 109 S6K1-pERK→ pS6K1
        v[109] = y[V.S6K1] * x[C.S6K1_phosphorylation_pERK] * y[V.pERK]
        # 109 pS6K1 −→ S6K1
        v[110] = x[C.pS6K1_dephosphorylation] * y[V.pS6K1]
        # 111 S6 -pS6K1→ pS6
        v[111] = y[V.S6] * x[C.S6_phosphorylation_pS6K1] * y[V.pS6K1]
        # 112 S6 pERK→ pS6
        v[112] = y[V.S6] * x[C.S6_phosphorylation_pERK] * y[V.pERK]
        # 113 pS6 −→ S6
        v[113] = x[C.pS6_dephosphorylation] * y[V.pS6]

        return v



from dataclasses import make_dataclass
from typing import Dict, List

NAMES: List[str] = [
    "AKT_activation_pEGFR",
    "AKT_activation_pErbB12",
    "AKT_activation_pErbB13",
    "AKT_activation_pErbB32",
    "AKT_activation_pIGF1R",
    "AKT_activation_pMetEGFR",
    "AKT_activation_pMetErbB3",
    "AKT_activation_pMetd",
    "AKT_internIGF1R_effect",
    "EGFR_BTC_binding",
    "EGFR_BTC_dimerize",
    "EGFR_ErbB2_BTC_dimerize",
    "EGFR_ErbB2_basal_act",
    "EGFR_ErbB2_dimerize",
    "EGFR_ErbB3_BTC_dimerize",
    "EGFR_ErbB3_basal_act",
    "EGFR_ErbB3_dimerize",
    "EGFR_ErbB3_dimerize_noHRG",
    "EGFR_basal_activation",
    "EGFR_basal_recycle",
    "EGFR_dimerize",
    "EGFR_lig_binding",
    "EGF_kD",
    "ERK_phosphorylation_pMEK",
    "ErbB2_ErbB3_dimerize",
    "ErbB2_dimerize",
    "ErbB2_recycle",
    "ErbB3_ErbB2_basal_act",
    "ErbB3_basal_activation",
    "ErbB3_basal_recycle",
    "ErbB3_dimerize",
    "ErbB3_lig_binding",
    "HGF_kD",
    "HRG_kD",
    "IGF1R_basal_activation",
    "IGF1R_basal_recycle",
    "IGF1R_dimerize",
    "IGF1R_lig_binding",
    "IGF1_kD",
    "MEK_internIGF1R_effect",
    "MEK_phosphorylation_pEGFR",
    "MEK_phosphorylation_pErbB12",
    "MEK_phosphorylation_pErbB13",
    "MEK_phosphorylation_pErbB32",
    "MEK_phosphorylation_pIGF1R",
    "MEK_phosphorylation_pMetEGFR",
    "MEK_phosphorylation_pMetErbB3",
    "MEK_phosphorylation_pMetd",
    "Met_EGFR_BTC_dimerize",
    "Met_EGFR_basal_act",
    "Met_EGFR_dimerize",
    "Met_ErbB3_basal_act",
    "Met_ErbB3_dimerize",
    "Met_basal_act",
    "Met_dimerize",
    "Met_lig_ErbB3_dimerize",
    "Met_lig_binding",
    "Met_recycle",
    "S6K1_phosphorylation_pAKT",
    "S6K1_phosphorylation_pERK",
    "S6_phosphorylation_pERK",
    "S6_phosphorylation_pS6K1",
    "feedback_pAKT",
    "feedback_pERK",
    "feedback_pERK_on_AKT",
    "feedback_pS6K1",
    "init_AKT",
    "init_EGFR",
    "init_EGFR_EGF",
    "init_ErbB2",
    "init_ErbB3",
    "init_ErbB3_HRG",
    "init_IGF1R",
    "init_IGF1R_IGF1",
    "init_MEK",
    "init_Met",
    "init_Met_HGF",
    "init_RTKph",
    "init_S6",
    "init_pERK",
    "init_pS6K1",
    "pAKT_deactivation",
    "pEGFR_degradation",
    "pEGFR_internalize",
    "pEGFR_phosphatase_binding",
    "pEGFRi_dephosph",
    "pERK_dephosphorylation",
    "pErbB12_degradation",
    "pErbB12_internalize",
    "pErbB12i_dephosph",
    "pErbB12i_phosphatase",
    "pErbB13_degradation",
    "pErbB13_internalize",
    "pErbB13i_dephosph",
    "pErbB13i_phosphatase",
    "pErbB2_degradation",
    "pErbB2_internalize",
    "pErbB2i_dephosph",
    "pErbB2i_phosphatase",
    "pErbB32_degradation",
    "pErbB32_internalize",
    "pErbB32i_dephosph",
    "pErbB32i_phosphatase",
    "pErbB3_degradation",
    "pErbB3_internalize",
    "pErbB3i_dephosph",
    "pErbB3i_phosphatase",
    "pIGF1R_degradation",
    "pIGF1R_internalize",
    "pIGF1Ri_dephosph",
    "pIGF1Ri_phosphatase",
    "pMEK_dephosphorylation",
    "pMetEGFR_degradation",
    "pMetEGFR_internalize",
    "pMetEGFRi_dephosph",
    "pMetEGFRi_phosphatase",
    "pMetErbB3_degradation",
    "pMetErbB3_internalize",
    "pMetErbB3i_dephosph",
    "pMetErbB3i_phosphatase",
    "pMet_degradation",
    "pMet_internalize",
    "pMeti_dephosph",
    "pMeti_phosphatase",
    "pS6K1_dephosphorylation",
    "pS6_dephosphorylation",
    "Ligand_on",
    "scale_Ligand",
    "ModelValue_122_0",
    "offset_pAKT_CelllineA431",
    "offset_pAKT_CelllineACHN_197",
    "offset_pAKT_CelllineACHN_200",
    "offset_pAKT_CelllineACHN_218",
    "offset_pAKT_CelllineACHN_DM",
    "offset_pAKT_CelllineADRr",
    "offset_pAKT_CelllineADRr_B",
    "offset_pAKT_CelllineADRr_B2",
    "offset_pAKT_CelllineBT20",
    "offset_pAKT_CelllineBxPc3",
    "offset_pAKT_CelllineH322M",
    "offset_pAKT_CelllineIGROV1",
    "offset_pEGFR_CelllineA431",
    "offset_pEGFR_CelllineACHN_197",
    "offset_pEGFR_CelllineACHN_200",
    "offset_pEGFR_CelllineACHN_218",
    "offset_pEGFR_CelllineACHN_DM",
    "offset_pEGFR_CelllineADRr",
    "offset_pEGFR_CelllineADRr_B",
    "offset_pEGFR_CelllineADRr_B2",
    "offset_pEGFR_CelllineBT20",
    "offset_pEGFR_CelllineBxPc3",
    "offset_pEGFR_CelllineH322M",
    "offset_pEGFR_CelllineIGROV1",
    "offset_pERK_CelllineA431",
    "offset_pERK_CelllineACHN_197",
    "offset_pERK_CelllineACHN_200",
    "offset_pERK_CelllineACHN_218",
    "offset_pERK_CelllineACHN_DM",
    "offset_pERK_CelllineADRr",
    "offset_pERK_CelllineADRr_B",
    "offset_pERK_CelllineADRr_B2",
    "offset_pERK_CelllineBT20",
    "offset_pERK_CelllineBxPc3",
    "offset_pERK_CelllineH322M",
    "offset_pERK_CelllineIGROV1",
    "offset_pErbB2_CelllineA431",
    "offset_pErbB2_CelllineACHN_197",
    "offset_pErbB2_CelllineACHN_200",
    "offset_pErbB2_CelllineACHN_218",
    "offset_pErbB2_CelllineACHN_DM",
    "offset_pErbB2_CelllineADRr",
    "offset_pErbB2_CelllineADRr_B",
    "offset_pErbB2_CelllineADRr_B2",
    "offset_pErbB2_CelllineBT20",
    "offset_pErbB2_CelllineBxPc3",
    "offset_pErbB2_CelllineH322M",
    "offset_pErbB2_CelllineIGROV1",
    "offset_pErbB3_CelllineA431",
    "offset_pErbB3_CelllineACHN_197",
    "offset_pErbB3_CelllineACHN_200",
    "offset_pErbB3_CelllineACHN_218",
    "offset_pErbB3_CelllineACHN_DM",
    "offset_pErbB3_CelllineADRr",
    "offset_pErbB3_CelllineADRr_B",
    "offset_pErbB3_CelllineADRr_B2",
    "offset_pErbB3_CelllineBT20",
    "offset_pErbB3_CelllineBxPc3",
    "offset_pErbB3_CelllineH322M",
    "offset_pErbB3_CelllineIGROV1",
    "offset_pIGF1R_CelllineA431",
    "offset_pIGF1R_CelllineACHN_197",
    "offset_pIGF1R_CelllineACHN_200",
    "offset_pIGF1R_CelllineACHN_218",
    "offset_pIGF1R_CelllineACHN_DM",
    "offset_pIGF1R_CelllineADRr",
    "offset_pIGF1R_CelllineADRr_B",
    "offset_pIGF1R_CelllineADRr_B2",
    "offset_pIGF1R_CelllineBT20",
    "offset_pIGF1R_CelllineBxPc3",
    "offset_pIGF1R_CelllineH322M",
    "offset_pIGF1R_CelllineIGROV1",
    "offset_pMEK_CelllineA431",
    "offset_pMEK_CelllineACHN_197",
    "offset_pMEK_CelllineACHN_200",
    "offset_pMEK_CelllineACHN_218",
    "offset_pMEK_CelllineACHN_DM",
    "offset_pMEK_CelllineADRr",
    "offset_pMEK_CelllineADRr_B",
    "offset_pMEK_CelllineADRr_B2",
    "offset_pMEK_CelllineBT20",
    "offset_pMEK_CelllineBxPc3",
    "offset_pMEK_CelllineH322M",
    "offset_pMEK_CelllineIGROV1",
    "offset_pMet_CelllineA431",
    "offset_pMet_CelllineACHN_197",
    "offset_pMet_CelllineACHN_200",
    "offset_pMet_CelllineACHN_218",
    "offset_pMet_CelllineACHN_DM",
    "offset_pMet_CelllineADRr",
    "offset_pMet_CelllineADRr_B",
    "offset_pMet_CelllineADRr_B2",
    "offset_pMet_CelllineBT20",
    "offset_pMet_CelllineBxPc3",
    "offset_pMet_CelllineH322M",
    "offset_pMet_CelllineIGROV1",
    "offset_pS6K1_CelllineA431",
    "offset_pS6K1_CelllineACHN_197",
    "offset_pS6K1_CelllineACHN_200",
    "offset_pS6K1_CelllineACHN_218",
    "offset_pS6K1_CelllineACHN_DM",
    "offset_pS6K1_CelllineADRr",
    "offset_pS6K1_CelllineADRr_B",
    "offset_pS6K1_CelllineADRr_B2",
    "offset_pS6K1_CelllineBT20",
    "offset_pS6K1_CelllineBxPc3",
    "offset_pS6K1_CelllineH322M",
    "offset_pS6K1_CelllineIGROV1",
    "offset_pS6_CelllineA431",
    "offset_pS6_CelllineACHN_197",
    "offset_pS6_CelllineACHN_200",
    "offset_pS6_CelllineACHN_218",
    "offset_pS6_CelllineACHN_DM",
    "offset_pS6_CelllineADRr",
    "offset_pS6_CelllineADRr_B",
    "offset_pS6_CelllineADRr_B2",
    "offset_pS6_CelllineBT20",
    "offset_pS6_CelllineBxPc3",
    "offset_pS6_CelllineH322M",
    "offset_pS6_CelllineIGROV1",
    "offset_tEGFR_CelllineA431",
    "offset_tEGFR_CelllineACHN_197",
    "offset_tEGFR_CelllineACHN_200",
    "offset_tEGFR_CelllineACHN_218",
    "offset_tEGFR_CelllineACHN_DM",
    "offset_tEGFR_CelllineADRr",
    "offset_tEGFR_CelllineADRr_B",
    "offset_tEGFR_CelllineADRr_B2",
    "offset_tEGFR_CelllineBT20",
    "offset_tEGFR_CelllineBxPc3",
    "offset_tEGFR_CelllineH322M",
    "offset_tEGFR_CelllineIGROV1",
    "offset_tErbB2_CelllineA431",
    "offset_tErbB2_CelllineACHN_197",
    "offset_tErbB2_CelllineACHN_200",
    "offset_tErbB2_CelllineACHN_218",
    "offset_tErbB2_CelllineACHN_DM",
    "offset_tErbB2_CelllineADRr",
    "offset_tErbB2_CelllineADRr_B",
    "offset_tErbB2_CelllineADRr_B2",
    "offset_tErbB2_CelllineBT20",
    "offset_tErbB2_CelllineBxPc3",
    "offset_tErbB2_CelllineH322M",
    "offset_tErbB2_CelllineIGROV1",
    "offset_tErbB3_CelllineA431",
    "offset_tErbB3_CelllineACHN_197",
    "offset_tErbB3_CelllineACHN_200",
    "offset_tErbB3_CelllineACHN_218",
    "offset_tErbB3_CelllineACHN_DM",
    "offset_tErbB3_CelllineADRr",
    "offset_tErbB3_CelllineADRr_B",
    "offset_tErbB3_CelllineADRr_B2",
    "offset_tErbB3_CelllineBT20",
    "offset_tErbB3_CelllineBxPc3",
    "offset_tErbB3_CelllineH322M",
    "offset_tErbB3_CelllineIGROV1",
    "offset_tIGF1R_CelllineA431",
    "offset_tIGF1R_CelllineACHN_197",
    "offset_tIGF1R_CelllineACHN_200",
    "offset_tIGF1R_CelllineACHN_218",
    "offset_tIGF1R_CelllineACHN_DM",
    "offset_tIGF1R_CelllineADRr",
    "offset_tIGF1R_CelllineADRr_B",
    "offset_tIGF1R_CelllineADRr_B2",
    "offset_tIGF1R_CelllineBT20",
    "offset_tIGF1R_CelllineBxPc3",
    "offset_tIGF1R_CelllineH322M",
    "offset_tIGF1R_CelllineIGROV1",
    "scale_pAKT_CelllineA431",
    "scale_pAKT_CelllineACHN_197",
    "scale_pAKT_CelllineACHN_200",
    "scale_pAKT_CelllineACHN_218",
    "scale_pAKT_CelllineACHN_DM",
    "scale_pAKT_CelllineADRr",
    "scale_pAKT_CelllineADRr_B",
    "scale_pAKT_CelllineADRr_B2",
    "scale_pAKT_CelllineBT20",
    "scale_pAKT_CelllineBxPc3",
    "scale_pAKT_CelllineH322M",
    "scale_pAKT_CelllineIGROV1",
    "scale_pEGFR_CelllineA431",
    "scale_pEGFR_CelllineACHN_197",
    "scale_pEGFR_CelllineACHN_200",
    "scale_pEGFR_CelllineACHN_218",
    "scale_pEGFR_CelllineACHN_DM",
    "scale_pEGFR_CelllineADRr",
    "scale_pEGFR_CelllineADRr_B",
    "scale_pEGFR_CelllineADRr_B2",
    "scale_pEGFR_CelllineBT20",
    "scale_pEGFR_CelllineBxPc3",
    "scale_pEGFR_CelllineH322M",
    "scale_pEGFR_CelllineIGROV1",
    "scale_pERK_CelllineA431",
    "scale_pERK_CelllineACHN_197",
    "scale_pERK_CelllineACHN_200",
    "scale_pERK_CelllineACHN_218",
    "scale_pERK_CelllineACHN_DM",
    "scale_pERK_CelllineADRr",
    "scale_pERK_CelllineADRr_B",
    "scale_pERK_CelllineADRr_B2",
    "scale_pERK_CelllineBT20",
    "scale_pERK_CelllineBxPc3",
    "scale_pERK_CelllineH322M",
    "scale_pERK_CelllineIGROV1",
    "scale_pErbB2_CelllineA431",
    "scale_pErbB2_CelllineACHN_197",
    "scale_pErbB2_CelllineACHN_200",
    "scale_pErbB2_CelllineACHN_218",
    "scale_pErbB2_CelllineACHN_DM",
    "scale_pErbB2_CelllineADRr",
    "scale_pErbB2_CelllineADRr_B",
    "scale_pErbB2_CelllineADRr_B2",
    "scale_pErbB2_CelllineBT20",
    "scale_pErbB2_CelllineBxPc3",
    "scale_pErbB2_CelllineH322M",
    "scale_pErbB2_CelllineIGROV1",
    "scale_pErbB3_CelllineA431",
    "scale_pErbB3_CelllineACHN_197",
    "scale_pErbB3_CelllineACHN_200",
    "scale_pErbB3_CelllineACHN_218",
    "scale_pErbB3_CelllineACHN_DM",
    "scale_pErbB3_CelllineADRr",
    "scale_pErbB3_CelllineADRr_B",
    "scale_pErbB3_CelllineADRr_B2",
    "scale_pErbB3_CelllineBT20",
    "scale_pErbB3_CelllineBxPc3",
    "scale_pErbB3_CelllineH322M",
    "scale_pErbB3_CelllineIGROV1",
    "scale_pIGF1R_CelllineA431",
    "scale_pIGF1R_CelllineACHN_197",
    "scale_pIGF1R_CelllineACHN_200",
    "scale_pIGF1R_CelllineACHN_218",
    "scale_pIGF1R_CelllineACHN_DM",
    "scale_pIGF1R_CelllineADRr",
    "scale_pIGF1R_CelllineADRr_B",
    "scale_pIGF1R_CelllineADRr_B2",
    "scale_pIGF1R_CelllineBT20",
    "scale_pIGF1R_CelllineBxPc3",
    "scale_pIGF1R_CelllineH322M",
    "scale_pIGF1R_CelllineIGROV1",
    "scale_pMEK_CelllineA431",
    "scale_pMEK_CelllineACHN_197",
    "scale_pMEK_CelllineACHN_200",
    "scale_pMEK_CelllineACHN_218",
    "scale_pMEK_CelllineACHN_DM",
    "scale_pMEK_CelllineADRr",
    "scale_pMEK_CelllineADRr_B",
    "scale_pMEK_CelllineADRr_B2",
    "scale_pMEK_CelllineBT20",
    "scale_pMEK_CelllineBxPc3",
    "scale_pMEK_CelllineH322M",
    "scale_pMEK_CelllineIGROV1",
    "scale_pMet_CelllineA431",
    "scale_pMet_CelllineACHN_197",
    "scale_pMet_CelllineACHN_200",
    "scale_pMet_CelllineACHN_218",
    "scale_pMet_CelllineACHN_DM",
    "scale_pMet_CelllineADRr",
    "scale_pMet_CelllineADRr_B",
    "scale_pMet_CelllineADRr_B2",
    "scale_pMet_CelllineBT20",
    "scale_pMet_CelllineBxPc3",
    "scale_pMet_CelllineH322M",
    "scale_pMet_CelllineIGROV1",
    "scale_pS6K1_CelllineA431",
    "scale_pS6K1_CelllineACHN_197",
    "scale_pS6K1_CelllineACHN_200",
    "scale_pS6K1_CelllineACHN_218",
    "scale_pS6K1_CelllineACHN_DM",
    "scale_pS6K1_CelllineADRr",
    "scale_pS6K1_CelllineADRr_B",
    "scale_pS6K1_CelllineADRr_B2",
    "scale_pS6K1_CelllineBT20",
    "scale_pS6K1_CelllineBxPc3",
    "scale_pS6K1_CelllineH322M",
    "scale_pS6K1_CelllineIGROV1",
    "scale_pS6_CelllineA431",
    "scale_pS6_CelllineACHN_197",
    "scale_pS6_CelllineACHN_200",
    "scale_pS6_CelllineACHN_218",
    "scale_pS6_CelllineACHN_DM",
    "scale_pS6_CelllineADRr",
    "scale_pS6_CelllineADRr_B",
    "scale_pS6_CelllineADRr_B2",
    "scale_pS6_CelllineBT20",
    "scale_pS6_CelllineBxPc3",
    "scale_pS6_CelllineH322M",
    "scale_pS6_CelllineIGROV1",
    "scale_tEGFR_CelllineA431",
    "scale_tEGFR_CelllineACHN_197",
    "scale_tEGFR_CelllineACHN_200",
    "scale_tEGFR_CelllineACHN_218",
    "scale_tEGFR_CelllineACHN_DM",
    "scale_tEGFR_CelllineADRr",
    "scale_tEGFR_CelllineADRr_B",
    "scale_tEGFR_CelllineADRr_B2",
    "scale_tEGFR_CelllineBT20",
    "scale_tEGFR_CelllineBxPc3",
    "scale_tEGFR_CelllineH322M",
    "scale_tEGFR_CelllineIGROV1",
    "scale_tErbB2_CelllineA431",
    "scale_tErbB2_CelllineACHN_197",
    "scale_tErbB2_CelllineACHN_200",
    "scale_tErbB2_CelllineACHN_218",
    "scale_tErbB2_CelllineACHN_DM",
    "scale_tErbB2_CelllineADRr",
    "scale_tErbB2_CelllineADRr_B",
    "scale_tErbB2_CelllineADRr_B2",
    "scale_tErbB2_CelllineBT20",
    "scale_tErbB2_CelllineBxPc3",
    "scale_tErbB2_CelllineH322M",
    "scale_tErbB2_CelllineIGROV1",
    "scale_tErbB3_CelllineA431",
    "scale_tErbB3_CelllineACHN_197",
    "scale_tErbB3_CelllineACHN_200",
    "scale_tErbB3_CelllineACHN_218",
    "scale_tErbB3_CelllineACHN_DM",
    "scale_tErbB3_CelllineADRr",
    "scale_tErbB3_CelllineADRr_B",
    "scale_tErbB3_CelllineADRr_B2",
    "scale_tErbB3_CelllineBT20",
    "scale_tErbB3_CelllineBxPc3",
    "scale_tErbB3_CelllineH322M",
    "scale_tErbB3_CelllineIGROV1",
    "scale_tIGF1R_CelllineA431",
    "scale_tIGF1R_CelllineACHN_197",
    "scale_tIGF1R_CelllineACHN_200",
    "scale_tIGF1R_CelllineACHN_218",
    "scale_tIGF1R_CelllineACHN_DM",
    "scale_tIGF1R_CelllineADRr",
    "scale_tIGF1R_CelllineADRr_B",
    "scale_tIGF1R_CelllineADRr_B2",
    "scale_tIGF1R_CelllineBT20",
    "scale_tIGF1R_CelllineBxPc3",
    "scale_tIGF1R_CelllineH322M",
    "scale_tIGF1R_CelllineIGROV1",
]

NUM: int = len(NAMES)

Parameters = make_dataclass(
    cls_name="Parameters",
    fields=[(name, int) for name in NAMES],
    namespace={"NAMES": NAMES, "NUM": NUM},
    frozen=True,
)

name2idx: Dict[str, int] = {k: v for v, k in enumerate(NAMES)}

C = Parameters(**name2idx)

del name2idx



import numpy as np

from biomass.dynamics.solver import solve_ode

from .name2idx import C, V
from .ode import DifferentialEquation


class Observable(DifferentialEquation):
    """
    Correlating model simulations and experimental measurements.

    Attributes
    ----------
    obs_names : list of strings
        Names of self.obs_names.

    t : range
        Simulation time span.

    conditions : list of strings
        Experimental conditions.

    simulations : numpy.ndarray
        The numpy array to store simulation results.

    normalization : nested dict
        * 'timepoint' : Optional[int]
            The time point at which simulated values are normalized.
            If :obj:`None`, the maximum value will be used for normalization.

        * 'condition' : list of strings
            The experimental conditions to use for normalization.
            If empty, all conditions defined in ``sim.conditions`` will be used.

    experiments : list of dict
        Time series data.

    error_bars : list of dict
        Error bars to show in figures.

    """

    def __init__(self):
        super(Observable, self).__init__(perturbation={})
        self.obs_names = [
            "pEGFR_au",
            "pErbB2_au",
            "pErbB3_au",
            "pERK_au",
            "pAKT_au",
            "pS6_au",
        ]
        self.t: range = range(241)
        self.conditions: list = ["EGF0nM", "EGF0_156nM", "EGF0_625nM", "EGF2_5nM", "EGF10nM"]
        self.simulations: np.ndarray = np.empty(
            (len(self.obs_names), len(self.conditions), len(self.t))
        )
        self.normalization: dict = {}
        self.experiments: list = [None] * len(self.obs_names)
        self.error_bars: list = [None] * len(self.obs_names)

    def simulate(self, x, y0, _perturbation=None):
        if _perturbation is not None:
            self.perturbation = _perturbation
        # add ligand
        for i, condition in enumerate(self.conditions):
            if condition == "EGF0nM":
                y0[V.dose_EGF] = 0
                y0[V.dose_HGF] = 0
                y0[V.dose_IGF1] = 0
                y0[V.dose_HRG] = 0
            elif condition == "EGF0_156nM":
                y0[V.dose_EGF] = 0.156 * x[C.scale_Ligand]
                y0[V.dose_IGF1] = 0
                y0[V.dose_HRG] = 0
            elif condition == "EGF0_625nM":
                y0[V.dose_EGF] = 0.625 * x[C.scale_Ligand]
                y0[V.dose_HGF] = 0
                y0[V.dose_IGF1] = 0
                y0[V.dose_HRG] = 0
            elif condition == "EGF2_5nM":
                y0[V.dose_EGF] = 2.5 * x[C.scale_Ligand]
                y0[V.dose_HGF] = 0
                y0[V.dose_IGF1] = 0
                y0[V.dose_HRG] = 0
            elif condition == "EGF10nM":
                y0[V.dose_EGF] = 10 * x[C.scale_Ligand]
                y0[V.dose_HGF] = 0
                y0[V.dose_IGF1] = 0
                y0[V.dose_HRG] = 0

            sol = solve_ode(self.diffeq, y0, self.t, tuple(x))

            if sol is None:
                return False
            else:
                self.simulations[self.obs_names.index("pEGFR_au"), i] = np.log10(
                    x[C.offset_pEGFR_CelllineH322M]
                    + x[C.scale_pEGFR_CelllineH322M]
                    * (
                        2 * sol.y[V.pEGFRd]
                        + 2 * sol.y[V.pEGFRi]
                        + 2 * sol.y[V.pEGFRi_ph]
                        + sol.y[V.pErbB12]
                        + sol.y[V.pErbB12i]
                        + sol.y[V.pErbB12i_ph]
                        + sol.y[V.pErbB13]
                        + sol.y[V.pErbB13i]
                        + sol.y[V.pErbB13i_ph]
                        + sol.y[V.pMetEGFR]
                        + sol.y[V.pMetEGFRi]
                        + sol.y[V.pMetEGFRi_ph]
                    )
                )
                self.simulations[self.obs_names.index("pErbB2_au"), i] = np.log10(
                    x[C.offset_pErbB2_CelllineH322M]
                    + x[C.scale_pErbB2_CelllineH322M]
                    * (
                        sol.y[V.pErbB12]
                        + sol.y[V.pErbB12i]
                        + sol.y[V.pErbB12i_ph]
                        + 2 * sol.y[V.pErbB2]
                        + 2 * sol.y[V.pErbB2i]
                        + 2 * sol.y[V.pErbB2i_ph]
                        + sol.y[V.pErbB32]
                        + sol.y[V.pErbB32i]
                        + sol.y[V.pErbB32i_ph]
                    )
                )
                self.simulations[self.obs_names.index("pErbB3_au"), i] = np.log10(
                    x[C.offset_pErbB3_CelllineH322M]
                    + x[C.scale_pErbB3_CelllineH322M]
                    * (
                        sol.y[V.pErbB13]
                        + sol.y[V.pErbB13i]
                        + sol.y[V.pErbB13i_ph]
                        + sol.y[V.pErbB32]
                        + sol.y[V.pErbB32i]
                        + sol.y[V.pErbB32i_ph]
                        + 2 * sol.y[V.pErbB3d]
                        + 2 * sol.y[V.pErbB3i]
                        + 2 * sol.y[V.pErbB3i_ph]
                        + sol.y[V.pMetErbB3]
                        + sol.y[V.pMetErbB3i]
                        + sol.y[V.pMetErbB3i_ph]
                    )
                )
                self.simulations[self.obs_names.index("pERK_au"), i] = np.log10(
                    x[C.offset_pERK_CelllineH322M]
                    + x[C.scale_pERK_CelllineH322M] * (sol.y[V.pERK])
                )
                self.simulations[self.obs_names.index("pAKT_au"), i] = np.log10(
                    x[C.offset_pAKT_CelllineH322M] + x[C.scale_pAKT_CelllineH322M] * sol.y[V.pAKT]
                )
                self.simulations[self.obs_names.index("pS6_au"), i] = np.log10(
                    x[C.offset_pS6_CelllineH322M] + x[C.scale_pS6_CelllineH322M] * sol.y[V.pS6]
                )
        return None

    def set_data(self):
        pass

    def get_timepoint(self, obs_name):
        if obs_name in self.obs_names:
            return []
        assert False



import os
import shutil

from biomass import Text2Model


def test_preprocessing():
    if os.path.isdir(
        os.path.join(
            os.path.dirname(__file__),
            "text_files",
            "abc",
        )
    ):
        shutil.rmtree(
            os.path.join(
                os.path.dirname(__file__),
                "text_files",
                "abc",
            )
        )


def test_text2model():
    abc = Text2Model(
        os.path.join(
            os.path.dirname(__file__),
            "text_files",
            "abc.txt",
        ),
    )
    abc.convert()
    # test thermodynamic restrictions
    desired = [["1", "2", "3", "4"]]
    actual = abc.restrictions
    assert len(actual) == 1
    assert set(desired[0]) == set(actual[0])


def test_cleanup():
    assert os.path.isdir(
        os.path.join(
            os.path.dirname(__file__),
            "text_files",
            "abc",
        )
    )
    shutil.rmtree(
        os.path.join(
            os.path.dirname(__file__),
            "text_files",
            "abc",
        )
    )



from dataclasses import make_dataclass
from typing import Dict, List

NAMES: List[str] = [
    "uRb",
    "tE2f",
    "E2f",
    "tE1",
    "tP21",
    "tCe",
    "tCa",
    "CeP21",
    "CaP21",
    "C1",
    "E1C1",
    "aPcna",
    "iPcna",
    "Rc",
    "pRc",
    "aRc",
    "iRc",
    "Dna",
    "P53",
    "Dam",
    "Pr",
]

NUM: int = len(NAMES)

Species = make_dataclass(
    cls_name="Species",
    fields=[(name, int) for name in NAMES],
    namespace={"NAMES": NAMES, "NUM": NUM},
    frozen=True,
)

name2idx: Dict[str, int] = {k: v for v, k in enumerate(NAMES)}

V = Species(**name2idx)

del name2idx



import multiprocessing
import os
import shutil
from distutils.dir_util import copy_tree
from typing import Final, Optional

import numpy as np
import pytest

from biomass import OptimizationResults, create_model, optimize, run_analysis, run_simulation
from biomass.estimation import InitialPopulation
from biomass.models import copy_to_current

MODEL_NAME: Final[str] = "Nakakuki_Cell_2010"

copy_to_current(MODEL_NAME)
assert os.path.exists(MODEL_NAME)
model = create_model(MODEL_NAME)


def test_initialization():
    os.mkdir(os.path.join(MODEL_NAME, "out"))
    copy_tree(
        os.path.join("tests", "out"),
        os.path.join(MODEL_NAME, "out"),
    )


def test_run_simulation():
    assert run_simulation(model, viz_type="experiment") is None
    with pytest.warns(RuntimeWarning) as record:
        run_simulation(model, viz_type="original")
    # check that the message matches
    assert record[-1].message.args[0] in [
        "Simulation failed. #original",
        "invalid value encountered in divide",
        "invalid value encountered in true_divide",
    ], f"got {record[-1].message.args[0]}"
    run_simulation(model, viz_type="average", stdev=True)
    assert os.path.isfile(
        os.path.join(
            model.path,
            "simulation_data",
            "simulations_all.npy",
        )
    )
    for obs_name in model.observables:
        assert os.path.isfile(
            os.path.join(model.path, "figure", "simulation", "average", f"{obs_name}.png")
        )


def test_save_resuts():
    res = OptimizationResults(model)
    res.savefig(figsize=(16, 5), boxplot_kws={"orient": "v"})
    assert os.path.isfile(
        os.path.join(
            model.path,
            "optimization_results",
            "estimated_parameter_sets.pdf",
        )
    )


def test_run_analysis():
    target = "reaction"
    metric = "integral"
    d_ln_vi = 0.009950330853168092
    run_analysis(model, target=target, metric=metric, options={"overwrite": True})
    layer = os.path.join("sensitivity_coefficients", f"{target}", f"{metric}.npy")
    loc_actual = os.path.join(model.path, layer)
    assert os.path.isfile(loc_actual)
    loc_desired = os.path.join(os.path.dirname(__file__), layer)
    actual = np.load(loc_actual)
    desired = np.load(loc_desired)
    assert np.shape(actual) == np.shape(desired)
    for i in range(actual.shape[0]):
        for j in range(actual.shape[1]):
            for k, obs_name in enumerate(model.observables):
                if obs_name != "Phosphorylated_MEKc":
                    np.allclose(
                        np.exp(actual[i, j, k] * d_ln_vi),
                        np.exp(desired[i, j, k] * d_ln_vi),
                        rtol=1e-3,
                    )


def test_param_estim(n_proc: Optional[int] = None):
    if n_proc is None:
        n_proc = max(1, multiprocessing.cpu_count())
    initpop = InitialPopulation(model).generate(n_proc=n_proc)
    optimize(model, x_id=11, optimizer_options={"maxiter": 5, "init": initpop, "workers": -1})
    with open(os.path.join(model.path, "out", "11", "optimization.log")) as f:
        logs = f.readlines()
    assert logs[-1].startswith("differential_evolution step 5: ")
    assert run_simulation(model, viz_type="11") is None
    assert os.path.isdir(os.path.join(model.path, "figure", "simulation", "11"))


def test_cleanup():
    shutil.rmtree(MODEL_NAME)



from dataclasses import make_dataclass
from typing import Dict, List

NAMES: List[str] = [
    "dose_EGF",
    "dose_HGF",
    "RTKph",
    "dose_IGF1",
    "dose_HRG",
    "dose_BTC",
    "EGFR",
    "EGFR_EGF",
    "EGFR_BTC",
    "pEGFRd",
    "pEGFRi",
    "pEGFRi_ph",
    "EGFRi",
    "ErbB2",
    "pErbB2",
    "pErbB2i",
    "ErbB2i",
    "pErbB2i_ph",
    "pErbB12",
    "pErbB12i",
    "pErbB12i_ph",
    "ErbB3",
    "ErbB3_HRG",
    "pErbB3d",
    "pErbB3i",
    "pErbB3i_ph",
    "ErbB3i",
    "pErbB13",
    "pErbB13i",
    "pErbB13i_ph",
    "pErbB32",
    "pErbB32i",
    "pErbB32i_ph",
    "IGF1R",
    "IGF1R_IGF1",
    "pIGF1Rd",
    "pIGF1Ri",
    "pIGF1Ri_ph",
    "IGF1Ri",
    "Met",
    "Met_HGF",
    "pMetd",
    "pMeti",
    "pMeti_ph",
    "Meti",
    "pMetErbB3",
    "pMetErbB3i",
    "pMetErbB3i_ph",
    "pMetEGFR",
    "pMetEGFRi",
    "pMetEGFRi_ph",
    "MEK",
    "pMEK",
    "ERK",
    "pERK",
    "AKT",
    "pAKT",
    "S6K1",
    "pS6K1",
    "S6",
    "pS6",
]

NUM: int = len(NAMES)

Species = make_dataclass(
    cls_name="Species",
    fields=[(name, int) for name in NAMES],
    namespace={"NAMES": NAMES, "NUM": NUM},
    frozen=True,
)

name2idx: Dict[str, int] = {k: v for v, k in enumerate(NAMES)}

V = Species(**name2idx)

del name2idx



from .name2idx import C, V
from .reaction_network import ReactionNetwork


class DifferentialEquation(ReactionNetwork):
    def __init__(self, perturbation):
        super(DifferentialEquation, self).__init__()
        self.perturbation = perturbation

    # Refined Model
    def diffeq(self, t, y, *x):
        v = self.flux(t, y, x)

        if self.perturbation:
            for i, dv in self.perturbation.items():
                v[i] = v[i] * dv

        dydt = [0] * V.NUM

        CycE = y[V.CycET] - y[V.CycEp27]
        CycA = y[V.CycAT] - y[V.CycAp27]

        Vdp27 = x[C.kd27] + (x[C.kd27e] * CycE + x[C.kd27a] * CycA) * y[V.Skp2]
        Vdcyce = (
            x[C.kdcyce]
            + x[C.kdcycee] * CycE / (1.0 + x[C.Inhibitor])
            + x[C.kdcycea] * CycA / (1.0 + x[C.Inhibitor])
        )
        Vdcyca = x[C.kdcyca] + x[C.kdcycac1] * y[V.Cdh1]
        Vdskp2 = x[C.kdskp2] + x[C.kdskp2c1] * y[V.Cdh1]

        Vicdh1 = x[C.kicdh1e] * CycE / (1.0 + x[C.Inhibitor]) + x[C.kicdh1a] * CycA / (
            1.0 + x[C.Inhibitor]
        )

        # protein
        dydt[V.p27T] = x[C.ks27] - Vdp27 * y[V.p27T]
        dydt[V.Skp2] = x[C.ksskp2] - Vdskp2 * y[V.Skp2]
        dydt[V.CycET] = x[C.kscyce] - Vdcyce * y[V.CycET]
        dydt[V.CycAT] = x[C.kscyca] - Vdcyca * y[V.CycAT]
        dydt[V.Emi1T] = x[C.ksemi1] - x[C.kdemi1] * y[V.Emi1T]

        dydt[V.CycEp27] = (
            x[C.kasse] * (y[V.CycET] - y[V.CycEp27]) * (y[V.p27T] - y[V.CycAp27] - y[V.CycEp27])
            - (x[C.kdise] + Vdp27 + Vdcyce) * y[V.CycEp27]
        )

        dydt[V.CycAp27] = (
            x[C.kassa] * (y[V.CycAT] - y[V.CycAp27]) * (y[V.p27T] - y[V.CycAp27] - y[V.CycEp27])
            - (x[C.kdisa] + Vdp27 + Vdcyca) * y[V.CycAp27]
        )

        dydt[V.EmiC] = (
            x[C.kasec] * (x[C.Cdh1T] - y[V.EmiC]) * (y[V.Emi1T] - y[V.EmiC])
            - (x[C.kdiec] + x[C.kdemi1]) * y[V.EmiC]
        )

        dydt[V.Cdh1dp] = x[C.kacdh1] * (x[C.Cdh1T] - y[V.Cdh1dp]) - Vicdh1 * y[V.Cdh1dp]

        dydt[V.Cdh1] = (
            (x[C.kdiec] + x[C.kdemi1]) * (y[V.Cdh1dp] - y[V.Cdh1])
            - x[C.kasec] * y[V.Cdh1] * (y[V.Emi1T] - y[V.EmiC])
            + x[C.kacdh1] * (x[C.Cdh1T] - y[V.EmiC] - y[V.Cdh1])
            - Vicdh1 * y[V.Cdh1]
        )

        return dydt


def param_values():
    x = [0] * C.NUM

    ## CYCE SYNTHESISx[C.DEGRADATION AND P27 BINDING/DISSOCIATION:
    x[C.kscyce] = 0.003
    x[C.kdcyce] = 0.001
    x[C.kdcycee] = 0.0001
    x[C.kdcycea] = 0.03
    x[C.kasse] = 1
    x[C.kdise] = 0.02
    ## CYCA SYNTHESISx[C.DEGRADATION AND P27 BINDING/DISSOCIATION:
    x[C.kscyca] = 0.0025
    x[C.kdcyca] = 0.002
    x[C.kdcycac1] = 0.4
    x[C.kassa] = 1
    x[C.kdisa] = 0.02
    ## P27 SYNTHESIS AND DEGRADATION:
    x[C.ks27] = 0.008
    x[C.kd27] = 0.004
    x[C.kd27e] = 2
    x[C.kd27a] = 2
    ## EMI1 SYNTHESIS AND DEGRADATION:
    x[C.ksemi1] = 0.003
    x[C.kdemi1] = 0.001
    ## CDH1 REGULATION:
    x[C.Cdh1T] = 1
    x[C.kacdh1] = 0.02
    x[C.kicdh1e] = 0.07
    x[C.kicdh1a] = 0.2
    x[C.kasec] = 2
    x[C.kdiec] = 0.02
    ## SKP2 SYNTHESIS AND DEGRADATION:
    x[C.ksskp2] = 0.004
    x[C.kdskp2] = 0.002
    x[C.kdskp2c1] = 0.2
    ## CDK INHIBITOR
    x[C.Inhibitor] = 0

    return x


def initial_values():
    y0 = [0] * V.NUM

    y0[V.Cdh1dp] = 1.0
    y0[V.Cdh1] = 1.0

    return y0



import multiprocessing
import os
import shutil
import sys
import warnings
from dataclasses import dataclass
from math import isfinite
from typing import Callable, List, Union

import numpy as np
from tqdm import tqdm

from ..model_object import ModelObject

DIRNAME = "_tmp"


class Tee(object):
    """
    Duplicate stdout to _tmp/optimization.log.
    """

    def __init__(self, model_path: str, x_id: int, disp_here: bool):
        self.disp_here = disp_here
        self.file = open(
            os.path.join(model_path, "out", DIRNAME + str(x_id), "optimization.log"),
            mode="w",
            encoding="utf-8",
        )
        self.stdout = sys.stdout
        sys.stdout = self

    def write(self, data):
        self.file.write(data)
        if self.disp_here:
            self.stdout.write(data)

    def __enter__(self):
        pass

    def __exit__(self, *args):
        sys.stdout = self.stdout
        self.file.close()


class Optimizer(object):
    """
    Use an external optimization method for parameterization of a mechanistic model.

    Attributes
    ----------
    model : ModelObject
        The BioMASS model object.
    optimize : Callable
        The optimizer, e.g., :func:`scipy.optimize.differential_evolution`.
    x_id : int
        Index of parameter set to estimate.
    disp_here : bool (default: :obj:`False`)
        Whether to show the evaluated *objective* at every iteration.
    overwrite : bool (default: :obj:`False`)
        If :obj:`True`, the directory (``x_id/``) will be overwritten.

    Examples
    --------
    >>> from scipy.optimize import differential_evolution
    >>> from biomass import create_model
    >>> from biomass.estimation import Optimizer
    >>> from biomass.models import copy_to_current
    >>> copy_to_current("Nakakuki_Cell_2010")
    >>> model = create_model("Nakakuki_Cell_2010")
    >>> param_idx = 1
    >>> optimizer = Optimizer(model, differential_evolution, param_idx)
    >>> def obj_fun(x):
    ...    '''Objective function to be minimized.'''
    ...    return model.get_obj_val(x)
    >>> res = optimizer.minimize(
    ...     obj_fun,
    ...     [(0, 1) for _ in range(len(model.problem.bounds))],
    ...     strategy="best1bin",
    ...     maxiter=50,
    ...     tol=1e-4,
    ...     mutation=0.1,
    ...     recombination=0.5,
    ...     disp=True,
    ...     polish=False,
    ...     workers=-1,
    ... )

    differential_evolution step 1: f(x)= 5.19392\n
    differential_evolution step 2: f(x)= 2.32477\n
    differential_evolution step 3: f(x)= 1.93583\n
    ...\n
    differential_evolution step 50: f(x)= 0.519774\n

    >>> from biomass import run_simulation
    >>> param_values = model.gene2val(res.x)
    >>> optimizer.import_solution(param_values)
    >>> run_simulation(model, viz_type=str(param_idx))
    """

    def __init__(
        self,
        model: ModelObject,
        optimize: Callable,
        x_id: int,
        disp_here: bool = False,
        overwrite: bool = False,
    ):
        self.model = model
        self.optimize = optimize
        self.x_id = x_id
        self.disp_here = disp_here

        self.savedir = os.path.join(self.model.path, "out", f"{self.x_id}")
        if os.path.isdir(self.savedir) and not overwrite:
            raise ValueError(
                f"out{os.sep}{self.x_id} already exists in {self.model.path}. "
                "Use another parameter id or set `overwrite` to True."
            )
        elif os.path.isdir(self.savedir) and overwrite:
            files = os.listdir(self.savedir)
            for file in files:
                if any(map(file.__contains__, (".npy", ".log"))):
                    os.remove(os.path.join(self.savedir, file))
        else:
            os.makedirs(self.savedir, exist_ok=True)
        os.makedirs(os.path.join(self.model.path, "out", DIRNAME + str(self.x_id)), exist_ok=True)
        self.default_stdout = sys.stdout

    def minimize(self, *args, **kwargs):
        """
        Execute the external optimizer.
        """
        os.makedirs(os.path.join(self.model.path, "out", DIRNAME + str(self.x_id)), exist_ok=True)
        with Tee(self.model.path, self.x_id, self.disp_here):
            with warnings.catch_warnings():
                warnings.simplefilter("ignore")
                res = self.optimize(*args, **kwargs)
        return res

    def _get_n_iter(self) -> int:
        n_iter: int = 0
        path_to_log = os.path.join(self.savedir, "optimization.log")
        with open(path_to_log, mode="r", encoding="utf-8") as f:
            log_file = f.readlines()
        for message in log_file:
            if len(message.strip()) > 0:
                n_iter += 1
        return n_iter

    def import_solution(self, x: Union[np.ndarray, List[float]], cleanup: bool = True) -> None:
        """
        Import the solution of the optimization to the model.
        The solution vector `x` will be saved to `path_to_model`/out/`x_id`/.
        Use ``biomass.run_simulation`` to visualize the optimization result.

        Parameters
        ----------
        x : Union[np.ndarray, List[float]]
            The solution of the optimization.
        cleanup : bool (default: True)
            If True (default), delete the temporary folder after the optimization is finished.
        """
        shutil.move(
            os.path.join(self.model.path, "out", DIRNAME + str(self.x_id), "optimization.log"),
            self.savedir,
        )

        best_fitness: float = self.model.problem.objective(x)
        n_iter = self._get_n_iter()
        np.save(os.path.join(self.savedir, "best_fitness"), best_fitness)
        np.save(os.path.join(self.savedir, "count_num"), n_iter)
        np.save(os.path.join(self.savedir, "generation"), n_iter)
        np.save(os.path.join(self.savedir, f"fit_param{n_iter}"), x)
        if cleanup:
            shutil.rmtree(os.path.join(self.model.path, "out", DIRNAME + str(self.x_id)))


@dataclass
class InitialPopulation(object):
    """
    Attributes
    ----------
    model : ModelObject
        BioMASS model object.
    popsize : int (default: 3)
        A multiplier for setting the total population size.
    threshold : float (default: 1e12)
        Allowable error for generating initial population.
        Default value is 1e12 (numerically solvable).
    """

    model: ModelObject
    popsize: int = 3
    threshold: float = 1e12

    def __post_init__(self) -> None:
        self.n_gene = len(self.model.problem.bounds)
        self.n_population = self.popsize * self.n_gene
        self.initpop_path = os.path.join(self.model.path, "_initpop")
        os.makedirs(self.initpop_path, exist_ok=True)

    def _set_gene_vector(self, pop_id: int) -> None:
        obj_val = np.inf
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            while self.threshold <= obj_val or not isfinite(obj_val):
                gene = np.random.rand(self.n_gene)
                obj_val = self.model.get_obj_val(gene)
            np.save(os.path.join(self.initpop_path, f"population_{pop_id}"), gene)

    def generate(self, n_proc: int = 1, progress: bool = False) -> np.ndarray:
        """
        Return initial population for ``optimizer_options['init']`` in ``biomass.optimize`` function.

        Parameters
        ----------
        n_proc : int (default: 1)
            Number of processes to use (default: 1). Set to a larger number
            (e.g. the number of CPU cores available) for parallel execution of simulations.
        progress : bool (default: :obj:`False`)
            Whether the progress bar is animating or not.

        Returns
        -------
        population : ``numpy.ndarray``
            Array specifying the initial population.
            The array should have shape (M, len(x)),
            where M is the total population size and len(x) is the number of parameters.

        Examples
        --------
        >>> from biomass import create_model, optimize
        >>> from biomass.estimation import InitialPopulation
        >>> from biomass.models import copy_to_current
        >>> copy_to_current("Nakakuki_Cell_2010")
        >>> model = create_model("Nakakuki_Cell_2010")
        >>> initpop = InitialPopulation(model).generate(n_proc=2, progress=True)
        >>> optimize(model, x_id=1, optimizer_options={"init": initpop})
        """
        p = multiprocessing.Pool(processes=n_proc)
        population = np.empty((self.n_population, self.n_gene))
        try:
            with tqdm(total=self.n_population, disable=not progress) as t:
                for _ in p.imap_unordered(self._set_gene_vector, range(self.n_population)):
                    t.update(1)
            for i in range(self.n_population):
                gene = np.load(os.path.join(self.initpop_path, f"population_{i}.npy"))
                population[i] = gene
            return population
        finally:
            p.close()
            shutil.rmtree(self.initpop_path)



from typing import Dict, List

from .name2idx import C, V


class ReactionNetwork(object):
    def __init__(self) -> None:
        """
        Reaction indices grouped according to biological processes.
        This is used for sensitivity analysis (target='reaction').
        """
        super(ReactionNetwork, self).__init__()

        self.reactions: Dict[str, List[int]] = {
            "ERK_activation": [i for i in range(1, 7)],
            "ERK_dephosphorylation_by_DUSP": [i for i in range(47, 57)],
            "ERK_transport": [i for i in range(7, 10)],
            "RSK_activation": [24, 25],
            "RSK_transport": [26],
            "Elk1_activation": [29, 30],
            "CREB_activation": [27, 28],
            "dusp_production_etc": [i for i in range(10, 14)],
            "DUSP_transport": [18, 19],
            "DUSP_stabilization": [14, 15, 20, 21],
            "DUSP_degradation": [16, 17, 22, 23],
            "cfos_production_etc": [i for i in range(31, 35)],
            "cFos_transport": [40, 41],
            "cFos_stabilization": [35, 36, 37, 42, 43, 44],
            "cFos_degradation": [38, 39, 45, 46],
            "Feedback_from_F": [i for i in range(57, 64)],
        }

    @staticmethod
    def flux(t, y, x) -> dict:
        """
        Rate equations in the model.

        Parameters
        ----------
        t : float
            Time point.
        y : ndarray
            Concentration vector.
        x : tuple
            Parameter values.

        Returns
        -------
        v : dict
            Flux vector.
        """
        v = {}
        # Rate equations
        v[1] = (
            x[C.V1]
            * x[C.a]
            * y[V.ppMEKc]
            * y[V.ERKc]
            / (x[C.Km1] * (1 + y[V.pERKc] / x[C.Km2]) + y[V.ERKc])
        )
        v[2] = (
            x[C.V2]
            * x[C.a]
            * y[V.ppMEKc]
            * y[V.pERKc]
            / (x[C.Km2] * (1 + y[V.ERKc] / x[C.Km1]) + y[V.pERKc])
        )
        v[3] = x[C.V3] * y[V.pERKc] / (x[C.Km3] * (1 + y[V.ppERKc] / x[C.Km4]) + y[V.pERKc])
        v[4] = x[C.V4] * y[V.ppERKc] / (x[C.Km4] * (1 + y[V.pERKc] / x[C.Km3]) + y[V.ppERKc])
        v[5] = x[C.V5] * y[V.pERKn] / (x[C.Km5] * (1 + y[V.ppERKn] / x[C.Km6]) + y[V.pERKn])
        v[6] = x[C.V6] * y[V.ppERKn] / (x[C.Km6] * (1 + y[V.pERKn] / x[C.Km5]) + y[V.ppERKn])
        v[7] = x[C.KimERK] * y[V.ERKc] - x[C.KexERK] * (x[C.Vn] / x[C.Vc]) * y[V.ERKn]
        v[8] = x[C.KimpERK] * y[V.pERKc] - x[C.KexpERK] * (x[C.Vn] / x[C.Vc]) * y[V.pERKn]
        v[9] = x[C.KimppERK] * y[V.ppERKc] - x[C.KexppERK] * (x[C.Vn] / x[C.Vc]) * y[V.ppERKn]
        v[10] = (
            x[C.V10] * y[V.ppERKn] ** x[C.n10] / (x[C.Km10] ** x[C.n10] + y[V.ppERKn] ** x[C.n10])
        )
        v[11] = x[C.p11] * y[V.PreduspmRNAn]
        v[12] = x[C.p12] * y[V.duspmRNAc]
        v[13] = x[C.p13] * y[V.duspmRNAc]
        v[14] = x[C.V14] * y[V.ppERKc] * y[V.DUSPc] / (x[C.Km14] + y[V.DUSPc])
        v[15] = x[C.V15] * y[V.pDUSPc] / (x[C.Km15] + y[V.pDUSPc])
        v[16] = x[C.p16] * y[V.DUSPc]
        v[17] = x[C.p17] * y[V.pDUSPc]
        v[18] = x[C.KimDUSP] * y[V.DUSPc] - x[C.KexDUSP] * (x[C.Vn] / x[C.Vc]) * y[V.DUSPn]
        v[19] = x[C.KimpDUSP] * y[V.pDUSPc] - x[C.KexpDUSP] * (x[C.Vn] / x[C.Vc]) * y[V.pDUSPn]
        v[20] = x[C.V20] * y[V.ppERKn] * y[V.DUSPn] / (x[C.Km20] + y[V.DUSPn])
        v[21] = x[C.V21] * y[V.pDUSPn] / (x[C.Km21] + y[V.pDUSPn])
        v[22] = x[C.p22] * y[V.DUSPn]
        v[23] = x[C.p23] * y[V.pDUSPn]
        v[24] = x[C.V24] * y[V.ppERKc] * y[V.RSKc] / (x[C.Km24] + y[V.RSKc])
        v[25] = x[C.V25] * y[V.pRSKc] / (x[C.Km25] + y[V.pRSKc])
        v[26] = x[C.KimRSK] * y[V.pRSKc] - x[C.KexRSK] * (x[C.Vn] / x[C.Vc]) * y[V.pRSKn]
        v[27] = x[C.V27] * y[V.pRSKn] * y[V.CREBn] / (x[C.Km27] + y[V.CREBn])
        v[28] = x[C.V28] * y[V.pCREBn] / (x[C.Km28] + y[V.pCREBn])
        v[29] = x[C.V29] * y[V.ppERKn] * y[V.Elk1n] / (x[C.Km29] + y[V.Elk1n])
        v[30] = x[C.V30] * y[V.pElk1n] / (x[C.Km30] + y[V.pElk1n])
        v[31] = (
            x[C.V31]
            * (y[V.pCREBn] * y[V.pElk1n]) ** x[C.n31]
            / (
                x[C.Km31] ** x[C.n31]
                + (y[V.pCREBn] * y[V.pElk1n]) ** x[C.n31]
                + (y[V.Fn] / x[C.KF31]) ** x[C.nF31]
            )
        )
        v[32] = x[C.p32] * y[V.PrecfosmRNAn]
        v[33] = x[C.p33] * y[V.cfosmRNAc]
        v[34] = x[C.p34] * y[V.cfosmRNAc]
        v[35] = x[C.V35] * y[V.ppERKc] * y[V.cFOSc] / (x[C.Km35] + y[V.cFOSc])
        v[36] = x[C.V36] * y[V.pRSKc] * y[V.cFOSc] / (x[C.Km36] + y[V.cFOSc])
        v[37] = x[C.V37] * y[V.pcFOSc] / (x[C.Km37] + y[V.pcFOSc])
        v[38] = x[C.p38] * y[V.cFOSc]
        v[39] = x[C.p39] * y[V.pcFOSc]
        v[40] = x[C.KimFOS] * y[V.cFOSc] - x[C.KexFOS] * (x[C.Vn] / x[C.Vc]) * y[V.cFOSn]
        v[41] = x[C.KimpcFOS] * y[V.pcFOSc] - x[C.KexpcFOS] * (x[C.Vn] / x[C.Vc]) * y[V.pcFOSn]
        v[42] = x[C.V42] * y[V.ppERKn] * y[V.cFOSn] / (x[C.Km42] + y[V.cFOSn])
        v[43] = x[C.V43] * y[V.pRSKn] * y[V.cFOSn] / (x[C.Km43] + y[V.cFOSn])
        v[44] = x[C.V44] * y[V.pcFOSn] / (x[C.Km44] + y[V.pcFOSn])
        v[45] = x[C.p45] * y[V.cFOSn]
        v[46] = x[C.p46] * y[V.pcFOSn]
        v[47] = x[C.p47] * y[V.DUSPn] * y[V.ppERKn] - x[C.m47] * y[V.DUSPn_ppERKn]
        v[48] = x[C.p48] * y[V.DUSPn_ppERKn]
        v[49] = x[C.p49] * y[V.DUSPn] * y[V.pERKn] - x[C.m49] * y[V.DUSPn_pERKn]
        v[50] = x[C.p50] * y[V.DUSPn_pERKn]
        v[51] = x[C.p51] * y[V.DUSPn] * y[V.ERKn] - x[C.m51] * y[V.DUSPn_ERKn]
        v[52] = x[C.p52] * y[V.pDUSPn] * y[V.ppERKn] - x[C.m52] * y[V.pDUSPn_ppERKn]
        v[53] = x[C.p53] * y[V.pDUSPn_ppERKn]
        v[54] = x[C.p54] * y[V.pDUSPn] * y[V.pERKn] - x[C.m54] * y[V.pDUSPn_pERKn]
        v[55] = x[C.p55] * y[V.pDUSPn_pERKn]
        v[56] = x[C.p56] * y[V.pDUSPn] * y[V.ERKn] - x[C.m56] * y[V.pDUSPn_ERKn]
        v[57] = (
            x[C.V57] * y[V.pcFOSn] ** x[C.n57] / (x[C.Km57] ** x[C.n57] + y[V.pcFOSn] ** x[C.n57])
        )
        v[58] = x[C.p58] * y[V.PreFmRNAn]
        v[59] = x[C.p59] * y[V.FmRNAc]
        v[60] = x[C.p60] * y[V.FmRNAc]
        v[61] = x[C.p61] * y[V.Fc]
        v[62] = x[C.KimF] * y[V.Fc] - x[C.KexF] * (x[C.Vn] / x[C.Vc]) * y[V.Fn]
        v[63] = x[C.p63] * y[V.Fn]

        return v



import os
import shutil

import matplotlib.pyplot as plt
import numpy as np

from biomass import create_model, run_simulation
from biomass.models import copy_to_current

MODEL_NAME: str = "g1s_transition"

copy_to_current(MODEL_NAME)
assert os.path.exists(MODEL_NAME)
model = create_model(MODEL_NAME)


def test_simulate_successful():
    x = model.pval()
    y0 = model.ival()
    assert model.problem.simulate(x, y0) is None


def test_run_simulation():
    assert run_simulation(model) is None
    res = np.load(os.path.join(model.path, "simulation_data", "simulations_original.npy"))

    plt.figure(figsize=(9, 6))
    plt.rcParams["font.family"] = "Arial"
    plt.rcParams["font.size"] = 16
    plt.rcParams["axes.linewidth"] = 1.5
    plt.rcParams["lines.linewidth"] = 2
    plt.rcParams["lines.markersize"] = 20

    plt.plot(model.problem.t, res[model.observables.index("p27_tot"), 0], "r-")
    plt.plot(model.problem.t, res[model.observables.index("CycE_tot"), 0], "b-")
    plt.plot(model.problem.t, res[model.observables.index("CycE"), 0], "b--")
    plt.plot(model.problem.t, res[model.observables.index("CycA_tot"), 0], "g-")
    plt.plot(model.problem.t, res[model.observables.index("CycA"), 0], "g--")

    plt.xticks([0, 300, 600, 900], [-5, 0, 5, 10])
    plt.xlim([0, 900])
    plt.xlabel("time relative to G1/S transition(h)")
    plt.yticks([0, 0.3, 0.6, 0.9])
    plt.ylabel("relative levels")

    plt.text(3.2 * 60, 0.75, "p27", ha="center", va="bottom", color="r")
    plt.text(6.4 * 60, 0.82, "CycE level", ha="center", va="bottom", color="b")
    plt.text(6.4 * 60, 0.55, "free\nCycE:Cdk2", ha="center", va="bottom", color="b")
    plt.text(12.4 * 60, 0.6, "CycA level", ha="center", va="bottom", color="g")
    plt.text(12.4 * 60, 0.35, "free\nCycA:Cdk2", ha="center", va="bottom", color="g")

    plt.show()


def test_cleanup():
    shutil.rmtree(MODEL_NAME)



from typing import Dict, List

from .name2idx import C, V


class ReactionNetwork(object):
    def __init__(self) -> None:
        super(ReactionNetwork, self).__init__()
        self.reactions: Dict[str, List[int]] = {}

    @staticmethod
    def flux(t, y, x):
        v = {}

        v[1] = x[C.k1_synthesis] * (y[V.pro_IRcom] - y[V.IRcom])
        v[2] = x[C.k1_InsIRcom] * y[V.Ins] * y[V.IRcom] - x[C.k2_InsIRcom] * y[V.p1IRcom]
        v[3] = x[C.k1_p1IRcomDeg] * y[V.p1IRcom]
        v[4] = x[C.k1_p1IRcomPhos] * y[V.pmTOR] * y[V.p1IRcom]
        v[5] = x[C.k1_p1p2IRcomdePhos] * y[V.p1p2IRcom]
        v[6] = x[C.k1_IRcomPhos] * y[V.pmTOR] * y[V.IRcom]
        v[7] = x[C.k1_p2IRcomdePhos] * y[V.p2IRcom]
        v[8] = x[C.k1_p2IRcomDeg] * y[V.p2IRcom]
        v[9] = x[C.k1_Insp2IRcom] * y[V.Ins] * y[V.p2IRcom] - x[C.k2_Insp2IRcom] * y[V.p1p2IRcom]
        v[10] = x[C.k1_p1p2IRcomDeg] * y[V.p1p2IRcom]
        v[11] = x[C.k1_AKTPhos] * (y[V.iAKT] - y[V.pAKT]) * y[V.p1IRcom]
        v[12] = x[C.k1_pAKTdePhos] * y[V.pAKT]
        v[13] = x[C.k1_mTORPhos] * (y[V.imTOR] - y[V.pmTOR]) * y[V.pAKT]
        v[14] = x[C.k1_pmTORdePhos] * y[V.pmTOR]
        v[15] = x[C.k1_S6KPhos] * (y[V.iS6K] - y[V.pS6K]) * y[V.pmTOR]
        v[16] = x[C.k1_pS6KdePhos] * y[V.pS6K] * y[V.pX]
        v[17] = x[C.k1_XPhos] * (y[V.iX] - y[V.pX]) * y[V.pmTOR]
        v[18] = x[C.k1_pXdePhos] * y[V.pX]
        v[19] = x[C.k1_GSK3BPhos] * (y[V.iGSK3B] - y[V.pGSK3B]) * y[V.pAKT]
        v[20] = x[C.k1_pGSK3BdePhos] * y[V.pGSK3B]
        v[21] = x[C.k1_FoxO1Phos] * (y[V.iFoxO1] - y[V.pFoxO1]) * y[V.pAKT]
        v[22] = x[C.k1_pFoxO1dePhos] * y[V.pFoxO1]
        v[23] = x[C.k1_G6PaseSynthesis] * (y[V.iFoxO1] - y[V.pFoxO1])
        v[24] = x[C.k1_G6PaseDeg] * y[V.G6Pase]

        return v



import os
import shutil
from typing import Callable

import numpy as np

from biomass import Text2Model, create_model, run_simulation


def test_preprocessing():
    if os.path.isdir(
        os.path.join(
            os.path.dirname(__file__),
            "text_files",
            "Kholodenko1999_2",
        )
    ):
        shutil.rmtree(
            os.path.join(
                os.path.dirname(__file__),
                "text_files",
                "Kholodenko1999_2",
            )
        )


def test_text2model():
    mapk_cascade = Text2Model(
        os.path.join(
            os.path.dirname(__file__),
            "text_files",
            "Kholodenko1999_2.txt",
        ),
    )
    mapk_cascade.convert()
    # test thermodynamic restrictions
    desired = [
        {"10", "11", "12", "9"},
        {"15", "17", "18", "21"},
        {"18", "19", "20", "22"},
        {"18", "19", "20", "22"},
        {"12", "17", "19", "24"},
        {"15", "20", "23", "24"},
        {"12", "21", "22", "23"},
    ]
    actual = [set(restriction) for restriction in mapk_cascade.restrictions]
    for l1 in actual:
        assert l1 in desired
    for l2 in desired:
        assert l2 in actual


def test_run_simulation():
    try:
        from .text_files import Kholodenko1999_2

        _packaging: Callable[[str], str] = lambda model_name: ".".join(
            ["tests.test_text2model.text_files", model_name]
        )
        model = create_model(_packaging("Kholodenko1999_2"))
        run_simulation(model)
        simulated_values = np.load(
            os.path.join(
                os.path.dirname(__file__),
                "text_files",
                "Kholodenko1999_2",
                "simulation_data",
                "simulations_original.npy",
            )
        )
        assert np.isfinite(simulated_values).all()
        expected = np.load(
            os.path.join(
                os.path.dirname(__file__),
                "simulations_original_BN.npy",
            )
        )
        assert np.allclose(simulated_values, expected)
    except ImportError as e:
        print(e)


def test_cleanup():
    if os.path.isdir(
        os.path.join(
            os.path.dirname(__file__),
            "text_files",
            "Kholodenko1999_2",
        )
    ):
        shutil.rmtree(
            os.path.join(
                os.path.dirname(__file__),
                "text_files",
                "Kholodenko1999_2",
            )
        )



import os
import re
from types import ModuleType
from typing import List, NamedTuple

import numpy as np


class OptimizedValues(NamedTuple):
    params: list
    initials: list


class ModelObject(object):
    """
    The BioMASS model object.

    Examples
    --------
    >>> from biomass import create_model
    >>> from biomass.models import mapk_cascade
    >>> model = create_model(mapk_cascade.__package__)
    >>> type(model)
    <class 'biomass.model_object.ModelObject'>
    >>> print('Parameters:', len(model.parameters))
    Parameters: 22
    >>> print('Species:', len(model.species))
    Species: 8
    >>> print('Observables:', len(model.observables))
    Observables: 2
    """

    def __init__(self, path: str, biomass_model: ModuleType):
        self._path = path
        self._parameters = biomass_model.C.NAMES
        self._species = biomass_model.V.NAMES
        self.pval = biomass_model.param_values
        self.ival = biomass_model.initial_values
        self.problem = biomass_model.OptimizationProblem()
        self.viz = biomass_model.Visualization()
        self.rxn = biomass_model.ReactionNetwork()

    @property
    def path(self) -> str:
        return self._path

    @property
    def parameters(self) -> List[str]:
        return self._parameters

    @property
    def species(self) -> list:
        return self._species

    @property
    def observables(self) -> List[str]:
        duplicate = [
            name for name in set(self.problem.obs_names) if self.problem.obs_names.count(name) > 1
        ]
        if not duplicate:
            return self.problem.obs_names
        else:
            raise NameError(f"Duplicate observables: {', '.join(duplicate)}")

    def get_individual(self, paramset_id: int) -> np.ndarray:
        """
        Get estimated parameter values from optimization results.

        Parameters
        ----------
        paramset_id : int
            Index of parameter set.

        Returns
        -------
        best_individual : ``numpy.ndarray``
            Estimated parameter values.
        """
        best_generation = np.load(
            os.path.join(
                self.path,
                "out",
                f"{paramset_id}",
                "generation.npy",
            )
        )
        best_individual = np.load(
            os.path.join(
                self.path,
                "out",
                f"{paramset_id}",
                f"fit_param{int(best_generation)}.npy",
            )
        )
        return best_individual

    def load_param(self, paramset: int) -> OptimizedValues:
        """
        Load a parameter set from optimization results.

        Parameters
        ----------
        paramset : int
            Index of parameter set.

        Returns
        -------
        optimized_values : :class:`biomass.model_object.OptimizedValues`
            Optimized parameter/initial values.
        """
        best_individual = self.get_individual(paramset)
        (x, y0) = self.problem.update(best_individual)
        optimized_values = OptimizedValues(x, y0)
        return optimized_values

    def get_executable(self) -> List[int]:
        """
        Get executable parameter sets from optimization results.
        """
        n_file = []
        try:
            fitparam_files = os.listdir(
                os.path.join(
                    self.path,
                    "out",
                )
            )
            for file in fitparam_files:
                if re.match(r"\d", file):
                    n_file.append(int(file))
            empty_folder = []
            for i, nth_paramset in enumerate(n_file):
                if not os.path.isfile(
                    os.path.join(
                        self.path,
                        "out",
                        f"{nth_paramset:d}",
                        "generation.npy",
                    )
                ):
                    empty_folder.append(i)
            for i in sorted(empty_folder, reverse=True):
                n_file.pop(i)
        except FileNotFoundError as e:
            print(e)
        return n_file

    def gene2val(self, indiv_gene: np.ndarray) -> np.ndarray:
        """
        Convert gene to actual value. Mainly used in parameter estimation.

        Parameters
        ----------
        indiv_gene : ``numpy.ndarray``
            Individual gene.

        Returns
        -------
        indiv_values : ``numpy.ndarray``
            Corresponding values.
        """
        bounds = self.problem.get_region()
        indiv_values = 10 ** (indiv_gene * (bounds[1, :] - bounds[0, :]) + bounds[0, :])
        return indiv_values

    def get_obj_val(self, indiv_gene: np.ndarray) -> float:
        """
        An objective function to minimize in parameter estimation.

        Parameters
        ----------
        indiv_gene : ``numpy.ndarray``
            Genes, not parameter values.

        Returns
        -------
        obj_val : float
            Objective function value.
        """
        obj_val = self.problem.objective(self.gene2val(indiv_gene))
        return obj_val



"""BioMASS - A Python Framework for Modeling and Analysis of Signaling Systems"""

from .construction import Text2Model
from .core import *
from .result import OptimizationResults
from .version import __version__

__author__ = __maintainer__ = "Hiroaki Imoto"
__email__ = "hiroaki.imoto@ucd.ie"



import numpy as np

from biomass.estimation import convert_scale, initialize_search_param

from .name2idx import C, V
from .ode import initial_values, param_values


class SearchParam(object):
    """Specify model parameters and/or initial values to optimize"""

    def __init__(self):
        # parameters
        self.idx_params = []

        # initial values
        self.idx_initials = [
            # V.(specie)
        ]

    def get_region(self):
        x = param_values()
        y0 = initial_values()

        search_param = initialize_search_param(
            parameters=C.NAMES,
            species=V.NAMES,
            param_values=x,
            initial_values=y0,
            estimated_params=self.idx_params,
            estimated_initials=self.idx_initials,
        )

        search_rgn = np.zeros((2, len(x) + len(y0)))
        # Default: 0.1 ~ 10
        for i, j in enumerate(self.idx_params):
            search_rgn[0, j] = search_param[i] * 0.1  # lower bound
            search_rgn[1, j] = search_param[i] * 10.0  # upper bound
        # Default: 0.5 ~ 2
        for i, j in enumerate(self.idx_initials):
            search_rgn[0, j + len(x)] = search_param[i + len(self.idx_params)] * 0.5  # lower bound
            search_rgn[1, j + len(x)] = search_param[i + len(self.idx_params)] * 2.0  # upper bound

        # search_rgn[:,C.parameter] = [lower_bound, upper_bound]
        # search_rgn[:,V.specie+len(x)] = [lower_bound, upper_bound]

        search_rgn = convert_scale(
            region=search_rgn,
            parameters=C.NAMES,
            species=V.NAMES,
            estimated_params=self.idx_params,
            estimated_initials=self.idx_initials,
        )

        return search_rgn

    def update(self, indiv):
        x = param_values()
        y0 = initial_values()

        for i, j in enumerate(self.idx_params):
            x[j] = indiv[i]
        for i, j in enumerate(self.idx_initials):
            y0[j] = indiv[i + len(self.idx_params)]

        return x, y0



from .name2idx import C, V
from .reaction_network import ReactionNetwork


class DifferentialEquation(ReactionNetwork):
    def __init__(self, perturbation):
        super(DifferentialEquation, self).__init__()
        self.perturbation = perturbation

    def diffeq(self, t, y, *x):
        """Kinetic equations"""
        v = self.flux(t, y, x)

        if self.perturbation:
            for i, dv in self.perturbation.items():
                v[i] = v[i] * dv

        dydt = [0] * V.NUM

        return dydt


def param_values():
    """Parameter values"""
    x = [1] * C.NUM

    return x


def initial_values():
    """Values of the initial condition"""
    y0 = [0] * V.NUM

    return y0



import os
import shutil

import pytest

from biomass import Text2Model
from biomass.construction.reaction_rules import DetectionError


def test_preprocessing():
    for i in ["1", "2"]:
        if os.path.isdir(
            os.path.join(
                os.path.dirname(__file__),
                "text_files",
                f"typo_{i}",
            )
        ):
            shutil.rmtree(
                os.path.join(
                    os.path.dirname(__file__),
                    "text_files",
                    f"typo_{i}",
                )
            )


def test_typo_detection():
    for i in ["1", "2"]:
        typo = Text2Model(
            os.path.join(
                os.path.dirname(__file__),
                "text_files",
                f"typo_{i}.txt",
            ),
        )
        with pytest.raises(DetectionError) as e:
            typo.convert()
        assert "Maybe: 'binds'." in str(e.value)


def test_unregistered_rule():
    ubiquitination = Text2Model(
        os.path.join(
            os.path.dirname(__file__),
            "text_files",
            "unregistered_rule.txt",
        )
    )
    with pytest.raises(DetectionError) as e:
        ubiquitination.convert()
    assert "Unregistered words in line1: A ubiquitinates B --> uB" in str(e.value)


def test_cleanup():
    for i in ["1", "2"]:
        assert os.path.isdir(
            os.path.join(
                os.path.dirname(__file__),
                "text_files",
                f"typo_{i}",
            )
        )
        shutil.rmtree(
            os.path.join(
                os.path.dirname(__file__),
                "text_files",
                f"typo_{i}",
            )
        )
    assert os.path.isdir(
        os.path.join(
            os.path.dirname(__file__),
            "text_files",
            "unregistered_rule",
        )
    )
    shutil.rmtree(
        os.path.join(
            os.path.dirname(__file__),
            "text_files",
            "unregistered_rule",
        )
    )



import matplotlib as mpl
import matplotlib.pyplot as plt

from biomass.plotting import *

from .observable import Observable


class Visualization(Observable):
    """
    Plotting parameters for customizing figure properties.

    Attributes
    ----------
    cm : matplotlib.colors.ListedColormap (default: ``plt.cm.get_cmap('tab10')``)
        Choosing colormaps for ``cmap``.
    single_observable_options : list of SingleObservable
        Visualization options for time-course simulation (single-observable).
    multiple_observables_options : MultipleObservables
        Visualization options for time-course simulation (multi-observables).
    sensitivity_options : SensitivityOptions
        Visualization options for sensitivity analysis results.
    """

    def __init__(self):
        super().__init__()

        self.cm = mpl.colormaps["tab10"]
        self.single_observable_options = [
            SingleObservable(self.cm, obs_name) for obs_name in self.obs_names
        ]
        self.multiple_observables_options = MultipleObservables(self.cm)
        self.sensitivity_options = SensitivityOptions(self.cm)

    def get_single_observable_options(self):
        for i, _ in enumerate(self.obs_names):
            self.single_observable_options[i].divided_by = 60
            self.single_observable_options[i].xlim = (0, 150)
            self.single_observable_options[i].xticks = [20 * i for i in range(8)]
            self.single_observable_options[i].xlabel = "TIME (min)"
            self.single_observable_options[i].ylim = (0, 310)
            self.single_observable_options[i].yticks = [50 * i for i in range(7)]
            self.single_observable_options[i].exp_data = False
            self.single_observable_options[
                self.obs_names.index("biphosphorylated_MAPK")
            ].ylabel = "ERK-PP"
            self.single_observable_options[
                self.obs_names.index("unphosphorylated_MAPK")
            ].ylabel = "ERK"

        return self.single_observable_options

    def get_multiple_observables_options(self):
        self.multiple_observables_options.fname = "oscillations_in_the_MAPK_concentrations"
        self.multiple_observables_options.observables = [
            "biphosphorylated_MAPK",
            "unphosphorylated_MAPK",
        ]
        self.multiple_observables_options.condition = "control"
        self.multiple_observables_options.xlim = (0, 150)
        self.multiple_observables_options.xticks = [20 * i for i in range(8)]
        self.multiple_observables_options.xlabel = "TIME (min)"
        self.multiple_observables_options.ylim = (0, 310)
        self.multiple_observables_options.yticks = [50 * i for i in range(7)]
        self.multiple_observables_options.ylabel = "MAPK concentrations (nM)"

        return self.multiple_observables_options

    def get_sensitivity_options(self):
        return self.sensitivity_options

    @staticmethod
    def set_timecourse_rcParams():
        """figure/simulation"""
        plt.rcParams["font.size"] = 12
        plt.rcParams["axes.linewidth"] = 1.5
        plt.rcParams["xtick.major.width"] = 1.5
        plt.rcParams["ytick.major.width"] = 1.5
        plt.rcParams["lines.linewidth"] = 1.8
        plt.rcParams["lines.markersize"] = 12
        plt.rcParams["savefig.bbox"] = "tight"
        # plt.rcParams["savefig.format"] = "pdf"
        # plt.rcParams['font.family'] = 'Arial'
        # plt.rcParams['mathtext.fontset'] = 'custom'
        # plt.rcParams['mathtext.it'] = 'Arial:italic'

    @staticmethod
    def set_sensitivity_rcParams():
        """figure/sensitivity"""
        plt.rcParams["font.size"] = 12
        plt.rcParams["axes.linewidth"] = 1.2
        plt.rcParams["xtick.major.width"] = 1.2
        plt.rcParams["ytick.major.width"] = 1.2
        plt.rcParams["savefig.bbox"] = "tight"
        # plt.rcParams["savefig.format"] = "pdf"
        # plt.rcParams['font.family'] = 'Arial'

    @staticmethod
    def convert_species_name(name):
        """figure/sensitivity/initial_condition
        - Sensitivity for species with nonzero initial conditions
        """
        """
        if name == 'MKKK':
            return 'Raf'
        elif name == 'MKK':
            return 'MEK'
        elif name == 'MAPK':
            return 'ERK'
        """
        return name



import os
import shutil

import matplotlib.pyplot as plt
import numpy as np

from biomass import create_model, run_simulation
from biomass.models import copy_to_current

MODEL_NAME: str = "pan_rtk"

copy_to_current(MODEL_NAME)
assert os.path.exists(MODEL_NAME)
model = create_model(MODEL_NAME)


def test_simulate_successful():
    x = model.pval()
    y0 = model.ival()
    assert model.problem.simulate(x, y0) is None


def test_run_simulation():
    assert run_simulation(model) is None
    res = np.load(os.path.join(model.path, "simulation_data", "simulations_original.npy"))

    colors = ["k", "b", "c", "r", "y"]
    yticks = [
        np.arange(0.5, 3, 0.5),
        np.arange(-0.8, 1.2, 0.2),
        np.arange(-0.8, 1, 0.2),
        np.arange(-1.2, 0.2, 0.2),
        np.arange(-0.8, 0.2, 0.2),
        np.arange(-0.5, 0.3, 0.1),
    ]
    sd = 1.0e-1

    plt.rcParams["font.size"] = 6
    plt.rcParams["font.family"] = "Arial"
    # plt.rcParams['axes.linewidth'] = 1
    plt.rcParams["lines.linewidth"] = 0.8

    plt.subplots_adjust(wspace=0.5, hspace=0.4)

    for i in range(6):
        if i < 3:
            plt.subplot(2, 4, i + 1)
        else:
            plt.subplot(2, 4, i + 2)
        plt.figsize = (7, 4)
        plt.gca().spines["right"].set_visible(False)
        plt.gca().spines["top"].set_visible(False)

        for j, color in enumerate(colors):
            plt.plot(
                model.problem.t,
                res[i, j],
                color,
                label=model.problem.conditions[j].replace("_", "."),
            )
            plt.fill_between(
                model.problem.t, res[i, j] - sd, res[i, j] + sd, facecolor=color, lw=0, alpha=0.1
            )
        plt.title(model.observables[i][:-3], fontweight="bold")
        plt.xlabel("time [min]")
        plt.xticks([0, 60, 120, 180, 240])
        plt.ylabel("(conc.) [au]")
        plt.yticks(yticks[i])
        if i == 2:
            plt.legend(bbox_to_anchor=(1.05, 1), loc="upper left", borderaxespad=0, frameon=False)
    plt.show()


def test_cleanup():
    shutil.rmtree(MODEL_NAME)



import os
import shutil

import matplotlib.pyplot as plt
import numpy as np

from biomass import create_model, run_simulation
from biomass.models import copy_to_current

MODEL_NAME: str = "nfkb_pathway"

copy_to_current(MODEL_NAME)
assert os.path.exists(MODEL_NAME)
model = create_model(MODEL_NAME)


def test_simulate_successful():
    x = model.pval()
    y0 = model.ival()
    assert model.problem.simulate(x, y0) is None


def test_example_plot():
    assert run_simulation(model, viz_type="original") is None
    res = np.load(os.path.join(model.path, "simulation_data", "simulations_original.npy"))

    fig = plt.figure(figsize=(9, 9))
    plt.rcParams["font.family"] = "Arial"
    plt.rcParams["font.size"] = 18
    plt.rcParams["axes.linewidth"] = 2
    plt.rcParams["lines.linewidth"] = 3
    plt.rcParams["lines.markersize"] = 16
    plt.subplots_adjust(wspace=0, hspace=0.3)

    for i, obs_name in enumerate(model.observables):
        plt.subplot(2, 1, i + 1)
        for j, (color, label) in enumerate(zip(["k", "r"], ["TNFα", "TNFα + DCF"])):
            plt.plot(model.problem.t, res[i, j], color=color, label=label)
        plt.title(f"{obs_name}".replace("_", " "))
        plt.xticks([0, 50, 100, 150, 200])
        if i == 0:
            plt.yticks([0, 0.05, 0.10, 0.15])
            plt.legend(loc="upper left", frameon=False)
    fig.text(0.5, 0.05, "time [min]", ha="center")
    fig.text(0.0, 0.5, "concentration [a.u.]", va="center", rotation="vertical")
    plt.show()


def test_cleanup():
    shutil.rmtree(MODEL_NAME)



from .name2idx import C, V
from .reaction_network import ReactionNetwork


class DifferentialEquation(ReactionNetwork):
    def __init__(self, perturbation):
        super(DifferentialEquation, self).__init__()
        self.perturbation = perturbation

    @staticmethod
    def _heaviside(x):
        return 1 * (x > 0)

    # Refined Model
    def diffeq(self, t, y, *x):
        v = self.flux(t, y, x)

        if self.perturbation:
            for i, dv in self.perturbation.items():
                v[i] = v[i] * dv

        Cd = 0.65
        tRb = 5
        tC1 = 1
        Skp2 = 1
        Cdt2 = 1

        ### ALGEBRAIC EQUATIONS
        Rb = y[V.uRb] - y[V.tE2f] + y[V.E2f]
        pRb = tRb - y[V.uRb]
        RbE2f = y[V.tE2f] - y[V.E2f]

        Ce = y[V.tCe] - y[V.CeP21]
        Ca = y[V.tCa] - y[V.CaP21]
        E1 = y[V.tE1] - y[V.E1C1]
        pC1 = tC1 - y[V.C1] - y[V.E1C1]
        P21 = y[V.tP21] - y[V.CeP21] - y[V.CaP21] - y[V.iPcna] - y[V.iRc]
        # tPcna = y[V.aPcna] + y[V.iPcna] + y[V.aRc] + y[V.iRc]

        ### REACTION RATES
        rSyE2f = x[C.kSyE2f] + x[C.kSyE2fE2f] * y[V.E2f] / (x[C.jSyE2f] + y[V.E2f])
        rPhRb = x[C.kPhRbCd] * Cd + x[C.kPhRbCe] * Ce + x[C.kPhRbCa] * Ca
        rPhRc = x[C.kPhRc] * (Ce + Ca) ** x[C.n] / (x[C.jCy] ** x[C.n] + (Ce + Ca) ** x[C.n])
        rPhC1 = x[C.kPhC1] + x[C.kPhC1Ce] * Ce + x[C.kPhC1Ca] * Ca

        rDeP21 = x[C.kDeP21] + x[C.kDeP21Cy] * Skp2 * (Ce + Ca) + x[C.kDeP21aRc] * Cdt2 * y[V.aRc]
        rDeCe = x[C.kDeCe] + x[C.kDeCeCa] * Ca
        rDeCa = x[C.kDeCa] + x[C.kDeCaC1] * y[V.C1]

        rDeP53 = x[C.kDeP53] / (x[C.jP53] + y[V.Dam])
        rReDam = x[C.kReDam] + x[C.kReDamP53] * y[V.P53] / (x[C.jDam] + y[V.Dam])

        rDsRc = self._heaviside(y[V.Dna] - 1)

        ### MODEL STATES
        dydt = [0] * V.NUM

        dydt[V.uRb] = -rPhRb * y[V.uRb] + x[C.kDpRb] * pRb
        dydt[V.tE2f] = rSyE2f - x[C.kDeE2f] * y[V.tE2f]
        dydt[V.E2f] = (
            rSyE2f
            - x[C.kDeE2f] * y[V.E2f]
            - x[C.kAsRbE2f] * Rb * y[V.E2f]
            + (x[C.kDsRbE2f] + rPhRb) * RbE2f
        )

        dydt[V.tE1] = x[C.kSyE1] * y[V.E2f] - x[C.kDeE1C1] * y[V.E1C1] - x[C.kDeE1] * E1
        dydt[V.tP21] = x[C.kSyP21] + x[C.kSyP21P53] * y[V.P53] - rDeP21 * y[V.tP21]
        dydt[V.tCe] = x[C.kSyCe] * y[V.E2f] - rDeCe * y[V.tCe]
        dydt[V.tCa] = x[C.kSyCa] * y[V.E2f] - rDeCa * y[V.tCa]

        dydt[V.CeP21] = x[C.kAsCyP21] * P21 * Ce - (x[C.kDsCyP21] + rDeCe + rDeP21) * y[V.CeP21]
        dydt[V.CaP21] = x[C.kAsCyP21] * P21 * Ca - (x[C.kDsCyP21] + rDeCa + rDeP21) * y[V.CaP21]

        dydt[V.C1] = (
            -rPhC1 * y[V.C1]
            + x[C.kDpC1] * pC1
            - x[C.kAsE1C1] * E1 * y[V.C1]
            + (x[C.kDsE1C1] + x[C.kDeE1C1]) * y[V.E1C1]
        )
        dydt[V.E1C1] = (
            -rPhC1 * y[V.E1C1]
            + x[C.kAsE1C1] * E1 * y[V.C1]
            - (x[C.kDsE1C1] + x[C.kDeE1C1]) * y[V.E1C1]
        )

        dydt[V.aPcna] = (
            x[C.kImPc]
            - x[C.kAsPcP21] * P21 * y[V.aPcna]
            + (x[C.kDsPcP21] + rDeP21) * y[V.iPcna]
            - (x[C.kExPc] + x[C.kAsRcPc] * y[V.pRc]) * y[V.aPcna]
            + (x[C.kDsRcPc] + rDsRc) * y[V.aRc]
        )
        dydt[V.iPcna] = (
            x[C.kAsPcP21] * P21 * y[V.aPcna]
            - (x[C.kDsPcP21] + rDeP21) * y[V.iPcna]
            - (x[C.kExPc] + x[C.kAsRcPc] * y[V.pRc]) * y[V.iPcna]
            + (x[C.kDsRcPc] + rDsRc) * y[V.iRc]
        )

        dydt[V.Rc] = -rPhRc * y[V.Rc] + x[C.kDpRc] * y[V.pRc] - rDsRc * y[V.Rc]
        dydt[V.pRc] = (
            rPhRc * y[V.Rc]
            - x[C.kDpRc] * y[V.pRc]
            - x[C.kAsRcPc] * (y[V.aPcna] + y[V.iPcna]) * y[V.pRc]
            + x[C.kDsRcPc] * (y[V.aRc] + y[V.iRc])
            - rDsRc * y[V.pRc]
        )
        dydt[V.aRc] = (
            -x[C.kAsPcP21] * P21 * y[V.aRc]
            + (x[C.kDsPcP21] + rDeP21) * y[V.iRc]
            + x[C.kAsRcPc] * y[V.aPcna] * y[V.pRc]
            - x[C.kDsRcPc] * y[V.aRc]
            - rDsRc * y[V.aRc]
        )
        dydt[V.iRc] = (
            x[C.kAsPcP21] * P21 * y[V.aRc]
            - (x[C.kDsPcP21] + rDeP21) * y[V.iRc]
            + x[C.kAsRcPc] * y[V.iPcna] * y[V.pRc]
            - x[C.kDsRcPc] * y[V.iRc]
            - rDsRc * y[V.iRc]
        )

        dydt[V.Dna] = x[C.kSyDna] * y[V.aRc]

        dydt[V.P53] = x[C.kSyP53] - rDeP53 * y[V.P53]
        dydt[V.Dam] = x[C.kGeDam] + x[C.kGeDamArc] * y[V.aRc] - rReDam * y[V.Dam]

        dydt[V.Pr] = x[C.kSyPr] - (x[C.kDePr] + x[C.kDeCaC1] * y[V.C1]) * y[V.Pr]

        return dydt


def param_values():
    x = [0] * C.NUM

    x[C.Bg] = 0.05
    x[C.kSyE2f] = 0.03
    x[C.kSyE2fE2f] = 0.04
    x[C.jSyE2f] = 0.2
    x[C.kAsRbE2f] = 5.0
    x[C.kDsRbE2f] = 0.005
    x[C.kDeE2f] = 0.05
    x[C.kPhRbCd] = 0.2
    x[C.kPhRbCe] = 0.3
    x[C.kPhRbCa] = 0.3
    x[C.kDpRb] = 0.05
    x[C.kSyE1] = 0.005
    x[C.kDeE1C1] = 0.005
    x[C.kDeE1] = 5.0e-4
    x[C.kPhC1] = 0.0
    x[C.kPhC1Ce] = 0.01
    x[C.kPhC1Ca] = 1.0
    x[C.kDpC1] = 0.05
    x[C.kAsE1C1] = 10.0
    x[C.kDsE1C1] = 0.01
    x[C.kSyP21] = 0.002
    x[C.kSyP21P53] = 0.008
    x[C.kDeP21] = 0.0025
    x[C.kDeP21Cy] = 0.007
    x[C.kDeP21aRc] = 1.0
    x[C.kSyCe] = 0.01
    x[C.kSyCa] = 0.02
    x[C.kAsCyP21] = 1.0
    x[C.kDsCyP21] = 0.05
    x[C.kDeCe] = 0.004
    x[C.kDeCa] = 0.01
    x[C.kDeCeCa] = 0.015
    x[C.kDeCaC1] = 2.0
    x[C.kImPc] = 0.003
    x[C.kExPc] = 0.006
    x[C.kPhRc] = 0.1
    x[C.kDpRc] = 0.05
    x[C.jCy] = 1.8
    x[C.n] = 6.0
    x[C.kAsRcPc] = 0.01
    x[C.kDsRcPc] = 0.001
    x[C.kAsPcP21] = 100.0
    x[C.kDsPcP21] = 0.01
    x[C.kSyDna] = 0.0093
    x[C.kSyP53] = 0.05
    x[C.kDeP53] = 0.05
    x[C.jP53] = 0.01
    x[C.kGeDam] = 0.001
    x[C.kGeDamArc] = 0.012
    x[C.kReDam] = 0.001
    x[C.kReDamP53] = 0.005
    x[C.jDam] = 0.5
    x[C.kSyPr] = 0.01
    x[C.kDePr] = 1.0e-4

    return x


def initial_values():
    y0 = [0] * V.NUM

    y0[V.tP21] = 0.6
    y0[V.aPcna] = 0.5
    y0[V.Rc] = 1
    y0[V.tCe] = 0.5
    y0[V.tCa] = 1.2
    y0[V.Pr] = 0.5

    return y0



from typing import Dict, List

from .name2idx import C, V


class ReactionNetwork(object):
    """
    Reaction indices grouped according to biological processes.
    This is used for sensitivity analysis (``target``='reaction').
    """

    def __init__(self) -> None:
        super(ReactionNetwork, self).__init__()
        self.reactions: Dict[str, List[int]] = {}

    @staticmethod
    def flux(t, y, x):
        """
        Flux vector.
        """

        v = {}

        return v
